# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all modules.
#
# CDP version: v0.0.1510116
# CDP domain: DOM

from __future__ import annotations

import enum
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any

import streamlink.webbrowser.cdp.devtools.page as page
import streamlink.webbrowser.cdp.devtools.runtime as runtime
from streamlink.webbrowser.cdp.devtools.util import T_JSON_DICT, CDPEvent


class NodeId(int):
    """
    Unique DOM node identifier.
    """
    def to_json(self) -> int:
        return self

    @classmethod
    def from_json(cls, json: int) -> NodeId:
        return cls(json)

    def __repr__(self):
        return f"NodeId({super().__repr__()})"


class BackendNodeId(int):
    """
    Unique DOM node identifier used to reference a node that may not have been pushed to the
    front-end.
    """
    def to_json(self) -> int:
        return self

    @classmethod
    def from_json(cls, json: int) -> BackendNodeId:
        return cls(json)

    def __repr__(self):
        return f"BackendNodeId({super().__repr__()})"


@dataclass
class BackendNode:
    """
    Backend node with a friendly name.
    """
    #: ``Node``'s nodeType.
    node_type: int

    #: ``Node``'s nodeName.
    node_name: str

    backend_node_id: BackendNodeId

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["nodeType"] = self.node_type
        json["nodeName"] = self.node_name
        json["backendNodeId"] = self.backend_node_id.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> BackendNode:
        return cls(
            node_type=int(json["nodeType"]),
            node_name=str(json["nodeName"]),
            backend_node_id=BackendNodeId.from_json(json["backendNodeId"]),
        )


class PseudoType(enum.Enum):
    """
    Pseudo element type.
    """
    FIRST_LINE = "first-line"
    FIRST_LETTER = "first-letter"
    CHECKMARK = "checkmark"
    BEFORE = "before"
    AFTER = "after"
    PICKER_ICON = "picker-icon"
    INTEREST_HINT = "interest-hint"
    MARKER = "marker"
    BACKDROP = "backdrop"
    COLUMN = "column"
    SELECTION = "selection"
    SEARCH_TEXT = "search-text"
    TARGET_TEXT = "target-text"
    SPELLING_ERROR = "spelling-error"
    GRAMMAR_ERROR = "grammar-error"
    HIGHLIGHT = "highlight"
    FIRST_LINE_INHERITED = "first-line-inherited"
    SCROLL_MARKER = "scroll-marker"
    SCROLL_MARKER_GROUP = "scroll-marker-group"
    SCROLL_BUTTON = "scroll-button"
    SCROLLBAR = "scrollbar"
    SCROLLBAR_THUMB = "scrollbar-thumb"
    SCROLLBAR_BUTTON = "scrollbar-button"
    SCROLLBAR_TRACK = "scrollbar-track"
    SCROLLBAR_TRACK_PIECE = "scrollbar-track-piece"
    SCROLLBAR_CORNER = "scrollbar-corner"
    RESIZER = "resizer"
    INPUT_LIST_BUTTON = "input-list-button"
    VIEW_TRANSITION = "view-transition"
    VIEW_TRANSITION_GROUP = "view-transition-group"
    VIEW_TRANSITION_IMAGE_PAIR = "view-transition-image-pair"
    VIEW_TRANSITION_GROUP_CHILDREN = "view-transition-group-children"
    VIEW_TRANSITION_OLD = "view-transition-old"
    VIEW_TRANSITION_NEW = "view-transition-new"
    PLACEHOLDER = "placeholder"
    FILE_SELECTOR_BUTTON = "file-selector-button"
    DETAILS_CONTENT = "details-content"
    PICKER = "picker"
    PERMISSION_ICON = "permission-icon"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> PseudoType:
        return cls(json)


class ShadowRootType(enum.Enum):
    """
    Shadow root type.
    """
    USER_AGENT = "user-agent"
    OPEN_ = "open"
    CLOSED = "closed"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> ShadowRootType:
        return cls(json)


class CompatibilityMode(enum.Enum):
    """
    Document compatibility mode.
    """
    QUIRKS_MODE = "QuirksMode"
    LIMITED_QUIRKS_MODE = "LimitedQuirksMode"
    NO_QUIRKS_MODE = "NoQuirksMode"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> CompatibilityMode:
        return cls(json)


class PhysicalAxes(enum.Enum):
    """
    ContainerSelector physical axes
    """
    HORIZONTAL = "Horizontal"
    VERTICAL = "Vertical"
    BOTH = "Both"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> PhysicalAxes:
        return cls(json)


class LogicalAxes(enum.Enum):
    """
    ContainerSelector logical axes
    """
    INLINE = "Inline"
    BLOCK = "Block"
    BOTH = "Both"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> LogicalAxes:
        return cls(json)


class ScrollOrientation(enum.Enum):
    """
    Physical scroll orientation
    """
    HORIZONTAL = "horizontal"
    VERTICAL = "vertical"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> ScrollOrientation:
        return cls(json)


@dataclass
class Node:
    """
    DOM interaction is implemented in terms of mirror objects that represent the actual DOM nodes.
    DOMNode is a base node mirror type.
    """
    #: Node identifier that is passed into the rest of the DOM messages as the ``nodeId``. Backend
    #: will only push node with given ``id`` once. It is aware of all requested nodes and will only
    #: fire DOM events for nodes known to the client.
    node_id: NodeId

    #: The BackendNodeId for this node.
    backend_node_id: BackendNodeId

    #: ``Node``'s nodeType.
    node_type: int

    #: ``Node``'s nodeName.
    node_name: str

    #: ``Node``'s localName.
    local_name: str

    #: ``Node``'s nodeValue.
    node_value: str

    #: The id of the parent node if any.
    parent_id: NodeId | None = None

    #: Child count for ``Container`` nodes.
    child_node_count: int | None = None

    #: Child nodes of this node when requested with children.
    children: list[Node] | None = None

    #: Attributes of the ``Element`` node in the form of flat array ``[name1, value1, name2, value2]``.
    attributes: list[str] | None = None

    #: Document URL that ``Document`` or ``FrameOwner`` node points to.
    document_url: str | None = None

    #: Base URL that ``Document`` or ``FrameOwner`` node uses for URL completion.
    base_url: str | None = None

    #: ``DocumentType``'s publicId.
    public_id: str | None = None

    #: ``DocumentType``'s systemId.
    system_id: str | None = None

    #: ``DocumentType``'s internalSubset.
    internal_subset: str | None = None

    #: ``Document``'s XML version in case of XML documents.
    xml_version: str | None = None

    #: ``Attr``'s name.
    name: str | None = None

    #: ``Attr``'s value.
    value: str | None = None

    #: Pseudo element type for this node.
    pseudo_type: PseudoType | None = None

    #: Pseudo element identifier for this node. Only present if there is a
    #: valid pseudoType.
    pseudo_identifier: str | None = None

    #: Shadow root type.
    shadow_root_type: ShadowRootType | None = None

    #: Frame ID for frame owner elements.
    frame_id: page.FrameId | None = None

    #: Content document for frame owner elements.
    content_document: Node | None = None

    #: Shadow root list for given element host.
    shadow_roots: list[Node] | None = None

    #: Content document fragment for template elements.
    template_content: Node | None = None

    #: Pseudo elements associated with this node.
    pseudo_elements: list[Node] | None = None

    #: Deprecated, as the HTML Imports API has been removed (crbug.com/937746).
    #: This property used to return the imported document for the HTMLImport links.
    #: The property is always undefined now.
    imported_document: Node | None = None

    #: Distributed nodes for given insertion point.
    distributed_nodes: list[BackendNode] | None = None

    #: Whether the node is SVG.
    is_svg: bool | None = None

    compatibility_mode: CompatibilityMode | None = None

    assigned_slot: BackendNode | None = None

    is_scrollable: bool | None = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["nodeId"] = self.node_id.to_json()
        json["backendNodeId"] = self.backend_node_id.to_json()
        json["nodeType"] = self.node_type
        json["nodeName"] = self.node_name
        json["localName"] = self.local_name
        json["nodeValue"] = self.node_value
        if self.parent_id is not None:
            json["parentId"] = self.parent_id.to_json()
        if self.child_node_count is not None:
            json["childNodeCount"] = self.child_node_count
        if self.children is not None:
            json["children"] = [i.to_json() for i in self.children]
        if self.attributes is not None:
            json["attributes"] = list(self.attributes)
        if self.document_url is not None:
            json["documentURL"] = self.document_url
        if self.base_url is not None:
            json["baseURL"] = self.base_url
        if self.public_id is not None:
            json["publicId"] = self.public_id
        if self.system_id is not None:
            json["systemId"] = self.system_id
        if self.internal_subset is not None:
            json["internalSubset"] = self.internal_subset
        if self.xml_version is not None:
            json["xmlVersion"] = self.xml_version
        if self.name is not None:
            json["name"] = self.name
        if self.value is not None:
            json["value"] = self.value
        if self.pseudo_type is not None:
            json["pseudoType"] = self.pseudo_type.to_json()
        if self.pseudo_identifier is not None:
            json["pseudoIdentifier"] = self.pseudo_identifier
        if self.shadow_root_type is not None:
            json["shadowRootType"] = self.shadow_root_type.to_json()
        if self.frame_id is not None:
            json["frameId"] = self.frame_id.to_json()
        if self.content_document is not None:
            json["contentDocument"] = self.content_document.to_json()
        if self.shadow_roots is not None:
            json["shadowRoots"] = [i.to_json() for i in self.shadow_roots]
        if self.template_content is not None:
            json["templateContent"] = self.template_content.to_json()
        if self.pseudo_elements is not None:
            json["pseudoElements"] = [i.to_json() for i in self.pseudo_elements]
        if self.imported_document is not None:
            json["importedDocument"] = self.imported_document.to_json()
        if self.distributed_nodes is not None:
            json["distributedNodes"] = [i.to_json() for i in self.distributed_nodes]
        if self.is_svg is not None:
            json["isSVG"] = self.is_svg
        if self.compatibility_mode is not None:
            json["compatibilityMode"] = self.compatibility_mode.to_json()
        if self.assigned_slot is not None:
            json["assignedSlot"] = self.assigned_slot.to_json()
        if self.is_scrollable is not None:
            json["isScrollable"] = self.is_scrollable
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Node:
        return cls(
            node_id=NodeId.from_json(json["nodeId"]),
            backend_node_id=BackendNodeId.from_json(json["backendNodeId"]),
            node_type=int(json["nodeType"]),
            node_name=str(json["nodeName"]),
            local_name=str(json["localName"]),
            node_value=str(json["nodeValue"]),
            parent_id=NodeId.from_json(json["parentId"]) if "parentId" in json else None,
            child_node_count=int(json["childNodeCount"]) if "childNodeCount" in json else None,
            children=[Node.from_json(i) for i in json["children"]] if "children" in json else None,
            attributes=[str(i) for i in json["attributes"]] if "attributes" in json else None,
            document_url=str(json["documentURL"]) if "documentURL" in json else None,
            base_url=str(json["baseURL"]) if "baseURL" in json else None,
            public_id=str(json["publicId"]) if "publicId" in json else None,
            system_id=str(json["systemId"]) if "systemId" in json else None,
            internal_subset=str(json["internalSubset"]) if "internalSubset" in json else None,
            xml_version=str(json["xmlVersion"]) if "xmlVersion" in json else None,
            name=str(json["name"]) if "name" in json else None,
            value=str(json["value"]) if "value" in json else None,
            pseudo_type=PseudoType.from_json(json["pseudoType"]) if "pseudoType" in json else None,
            pseudo_identifier=str(json["pseudoIdentifier"]) if "pseudoIdentifier" in json else None,
            shadow_root_type=ShadowRootType.from_json(json["shadowRootType"]) if "shadowRootType" in json else None,
            frame_id=page.FrameId.from_json(json["frameId"]) if "frameId" in json else None,
            content_document=Node.from_json(json["contentDocument"]) if "contentDocument" in json else None,
            shadow_roots=[Node.from_json(i) for i in json["shadowRoots"]] if "shadowRoots" in json else None,
            template_content=Node.from_json(json["templateContent"]) if "templateContent" in json else None,
            pseudo_elements=[Node.from_json(i) for i in json["pseudoElements"]] if "pseudoElements" in json else None,
            imported_document=Node.from_json(json["importedDocument"]) if "importedDocument" in json else None,
            distributed_nodes=[BackendNode.from_json(i) for i in json["distributedNodes"]] if "distributedNodes" in json else None,
            is_svg=bool(json["isSVG"]) if "isSVG" in json else None,
            compatibility_mode=CompatibilityMode.from_json(json["compatibilityMode"]) if "compatibilityMode" in json else None,
            assigned_slot=BackendNode.from_json(json["assignedSlot"]) if "assignedSlot" in json else None,
            is_scrollable=bool(json["isScrollable"]) if "isScrollable" in json else None,
        )


@dataclass
class DetachedElementInfo:
    """
    A structure to hold the top-level node of a detached tree and an array of its retained descendants.
    """
    tree_node: Node

    retained_node_ids: list[NodeId]

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["treeNode"] = self.tree_node.to_json()
        json["retainedNodeIds"] = [i.to_json() for i in self.retained_node_ids]
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> DetachedElementInfo:
        return cls(
            tree_node=Node.from_json(json["treeNode"]),
            retained_node_ids=[NodeId.from_json(i) for i in json["retainedNodeIds"]],
        )


@dataclass
class RGBA:
    """
    A structure holding an RGBA color.
    """
    #: The red component, in the [0-255] range.
    r: int

    #: The green component, in the [0-255] range.
    g: int

    #: The blue component, in the [0-255] range.
    b: int

    #: The alpha component, in the [0-1] range (default: 1).
    a: float | None = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["r"] = self.r
        json["g"] = self.g
        json["b"] = self.b
        if self.a is not None:
            json["a"] = self.a
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> RGBA:
        return cls(
            r=int(json["r"]),
            g=int(json["g"]),
            b=int(json["b"]),
            a=float(json["a"]) if "a" in json else None,
        )


class Quad(list):
    """
    An array of quad vertices, x immediately followed by y for each point, points clock-wise.
    """
    def to_json(self) -> list[float]:
        return self

    @classmethod
    def from_json(cls, json: list[float]) -> Quad:
        return cls(json)

    def __repr__(self):
        return f"Quad({super().__repr__()})"


@dataclass
class BoxModel:
    """
    Box model.
    """
    #: Content box
    content: Quad

    #: Padding box
    padding: Quad

    #: Border box
    border: Quad

    #: Margin box
    margin: Quad

    #: Node width
    width: int

    #: Node height
    height: int

    #: Shape outside coordinates
    shape_outside: ShapeOutsideInfo | None = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["content"] = self.content.to_json()
        json["padding"] = self.padding.to_json()
        json["border"] = self.border.to_json()
        json["margin"] = self.margin.to_json()
        json["width"] = self.width
        json["height"] = self.height
        if self.shape_outside is not None:
            json["shapeOutside"] = self.shape_outside.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> BoxModel:
        return cls(
            content=Quad.from_json(json["content"]),
            padding=Quad.from_json(json["padding"]),
            border=Quad.from_json(json["border"]),
            margin=Quad.from_json(json["margin"]),
            width=int(json["width"]),
            height=int(json["height"]),
            shape_outside=ShapeOutsideInfo.from_json(json["shapeOutside"]) if "shapeOutside" in json else None,
        )


@dataclass
class ShapeOutsideInfo:
    """
    CSS Shape Outside details.
    """
    #: Shape bounds
    bounds: Quad

    #: Shape coordinate details
    shape: list[Any]

    #: Margin shape bounds
    margin_shape: list[Any]

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["bounds"] = self.bounds.to_json()
        json["shape"] = list(self.shape)
        json["marginShape"] = list(self.margin_shape)
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ShapeOutsideInfo:
        return cls(
            bounds=Quad.from_json(json["bounds"]),
            shape=list(json["shape"]),
            margin_shape=list(json["marginShape"]),
        )


@dataclass
class Rect:
    """
    Rectangle.
    """
    #: X coordinate
    x: float

    #: Y coordinate
    y: float

    #: Rectangle width
    width: float

    #: Rectangle height
    height: float

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["x"] = self.x
        json["y"] = self.y
        json["width"] = self.width
        json["height"] = self.height
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Rect:
        return cls(
            x=float(json["x"]),
            y=float(json["y"]),
            width=float(json["width"]),
            height=float(json["height"]),
        )


@dataclass
class CSSComputedStyleProperty:
    #: Computed style property name.
    name: str

    #: Computed style property value.
    value: str

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["name"] = self.name
        json["value"] = self.value
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> CSSComputedStyleProperty:
        return cls(
            name=str(json["name"]),
            value=str(json["value"]),
        )


def collect_class_names_from_subtree(
    node_id: NodeId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[str]]:
    """
    Collects class names for the node with given id and all of it's child nodes.

    **EXPERIMENTAL**

    :param node_id: Id of the node to collect class names.
    :returns: Class name list.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.collectClassNamesFromSubtree",
        "params": params,
    }
    json = yield cmd_dict
    return [str(i) for i in json["classNames"]]


def copy_to(
    node_id: NodeId,
    target_node_id: NodeId,
    insert_before_node_id: NodeId | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, NodeId]:
    """
    Creates a deep copy of the specified node and places it into the target container before the
    given anchor.

    **EXPERIMENTAL**

    :param node_id: Id of the node to copy.
    :param target_node_id: Id of the element to drop the copy into.
    :param insert_before_node_id: *(Optional)* Drop the copy before this node (if absent, the copy becomes the last child of ```targetNodeId```).
    :returns: Id of the node clone.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["targetNodeId"] = target_node_id.to_json()
    if insert_before_node_id is not None:
        params["insertBeforeNodeId"] = insert_before_node_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.copyTo",
        "params": params,
    }
    json = yield cmd_dict
    return NodeId.from_json(json["nodeId"])


def describe_node(
    node_id: NodeId | None = None,
    backend_node_id: BackendNodeId | None = None,
    object_id: runtime.RemoteObjectId | None = None,
    depth: int | None = None,
    pierce: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, Node]:
    """
    Describes node given its id, does not require domain to be enabled. Does not start tracking any
    objects, can be used for automation.

    :param node_id: *(Optional)* Identifier of the node.
    :param backend_node_id: *(Optional)* Identifier of the backend node.
    :param object_id: *(Optional)* JavaScript object id of the node wrapper.
    :param depth: *(Optional)* The maximum depth at which children should be retrieved, defaults to 1. Use -1 for the entire subtree or provide an integer larger than 0.
    :param pierce: *(Optional)* Whether or not iframes and shadow roots should be traversed when returning the subtree (default is false).
    :returns: Node description.
    """
    params: T_JSON_DICT = {}
    if node_id is not None:
        params["nodeId"] = node_id.to_json()
    if backend_node_id is not None:
        params["backendNodeId"] = backend_node_id.to_json()
    if object_id is not None:
        params["objectId"] = object_id.to_json()
    if depth is not None:
        params["depth"] = depth
    if pierce is not None:
        params["pierce"] = pierce
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.describeNode",
        "params": params,
    }
    json = yield cmd_dict
    return Node.from_json(json["node"])


def scroll_into_view_if_needed(
    node_id: NodeId | None = None,
    backend_node_id: BackendNodeId | None = None,
    object_id: runtime.RemoteObjectId | None = None,
    rect: Rect | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Scrolls the specified rect of the given node into view if not already visible.
    Note: exactly one between nodeId, backendNodeId and objectId should be passed
    to identify the node.

    :param node_id: *(Optional)* Identifier of the node.
    :param backend_node_id: *(Optional)* Identifier of the backend node.
    :param object_id: *(Optional)* JavaScript object id of the node wrapper.
    :param rect: *(Optional)* The rect to be scrolled into view, relative to the node's border box, in CSS pixels. When omitted, center of the node will be used, similar to Element.scrollIntoView.
    """
    params: T_JSON_DICT = {}
    if node_id is not None:
        params["nodeId"] = node_id.to_json()
    if backend_node_id is not None:
        params["backendNodeId"] = backend_node_id.to_json()
    if object_id is not None:
        params["objectId"] = object_id.to_json()
    if rect is not None:
        params["rect"] = rect.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.scrollIntoViewIfNeeded",
        "params": params,
    }
    yield cmd_dict


def disable() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Disables DOM agent for the given page.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.disable",
    }
    yield cmd_dict


def discard_search_results(
    search_id: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Discards search results from the session with the given id. ``getSearchResults`` should no longer
    be called for that search.

    **EXPERIMENTAL**

    :param search_id: Unique search session identifier.
    """
    params: T_JSON_DICT = {}
    params["searchId"] = search_id
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.discardSearchResults",
        "params": params,
    }
    yield cmd_dict


def enable(
    include_whitespace: str | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enables DOM agent for the given page.

    :param include_whitespace: **(EXPERIMENTAL)** *(Optional)* Whether to include whitespaces in the children array of returned Nodes.
    """
    params: T_JSON_DICT = {}
    if include_whitespace is not None:
        params["includeWhitespace"] = include_whitespace
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.enable",
        "params": params,
    }
    yield cmd_dict


def focus(
    node_id: NodeId | None = None,
    backend_node_id: BackendNodeId | None = None,
    object_id: runtime.RemoteObjectId | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Focuses the given element.

    :param node_id: *(Optional)* Identifier of the node.
    :param backend_node_id: *(Optional)* Identifier of the backend node.
    :param object_id: *(Optional)* JavaScript object id of the node wrapper.
    """
    params: T_JSON_DICT = {}
    if node_id is not None:
        params["nodeId"] = node_id.to_json()
    if backend_node_id is not None:
        params["backendNodeId"] = backend_node_id.to_json()
    if object_id is not None:
        params["objectId"] = object_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.focus",
        "params": params,
    }
    yield cmd_dict


def get_attributes(
    node_id: NodeId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[str]]:
    """
    Returns attributes for the specified node.

    :param node_id: Id of the node to retrieve attributes for.
    :returns: An interleaved array of node attribute names and values.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getAttributes",
        "params": params,
    }
    json = yield cmd_dict
    return [str(i) for i in json["attributes"]]


def get_box_model(
    node_id: NodeId | None = None,
    backend_node_id: BackendNodeId | None = None,
    object_id: runtime.RemoteObjectId | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, BoxModel]:
    """
    Returns boxes for the given node.

    :param node_id: *(Optional)* Identifier of the node.
    :param backend_node_id: *(Optional)* Identifier of the backend node.
    :param object_id: *(Optional)* JavaScript object id of the node wrapper.
    :returns: Box model for the node.
    """
    params: T_JSON_DICT = {}
    if node_id is not None:
        params["nodeId"] = node_id.to_json()
    if backend_node_id is not None:
        params["backendNodeId"] = backend_node_id.to_json()
    if object_id is not None:
        params["objectId"] = object_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getBoxModel",
        "params": params,
    }
    json = yield cmd_dict
    return BoxModel.from_json(json["model"])


def get_content_quads(
    node_id: NodeId | None = None,
    backend_node_id: BackendNodeId | None = None,
    object_id: runtime.RemoteObjectId | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[Quad]]:
    """
    Returns quads that describe node position on the page. This method
    might return multiple quads for inline nodes.

    **EXPERIMENTAL**

    :param node_id: *(Optional)* Identifier of the node.
    :param backend_node_id: *(Optional)* Identifier of the backend node.
    :param object_id: *(Optional)* JavaScript object id of the node wrapper.
    :returns: Quads that describe node layout relative to viewport.
    """
    params: T_JSON_DICT = {}
    if node_id is not None:
        params["nodeId"] = node_id.to_json()
    if backend_node_id is not None:
        params["backendNodeId"] = backend_node_id.to_json()
    if object_id is not None:
        params["objectId"] = object_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getContentQuads",
        "params": params,
    }
    json = yield cmd_dict
    return [Quad.from_json(i) for i in json["quads"]]


def get_document(
    depth: int | None = None,
    pierce: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, Node]:
    """
    Returns the root DOM node (and optionally the subtree) to the caller.
    Implicitly enables the DOM domain events for the current target.

    :param depth: *(Optional)* The maximum depth at which children should be retrieved, defaults to 1. Use -1 for the entire subtree or provide an integer larger than 0.
    :param pierce: *(Optional)* Whether or not iframes and shadow roots should be traversed when returning the subtree (default is false).
    :returns: Resulting node.
    """
    params: T_JSON_DICT = {}
    if depth is not None:
        params["depth"] = depth
    if pierce is not None:
        params["pierce"] = pierce
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getDocument",
        "params": params,
    }
    json = yield cmd_dict
    return Node.from_json(json["root"])


def get_flattened_document(
    depth: int | None = None,
    pierce: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[Node]]:
    """
    Returns the root DOM node (and optionally the subtree) to the caller.
    Deprecated, as it is not designed to work well with the rest of the DOM agent.
    Use DOMSnapshot.captureSnapshot instead.

    :param depth: *(Optional)* The maximum depth at which children should be retrieved, defaults to 1. Use -1 for the entire subtree or provide an integer larger than 0.
    :param pierce: *(Optional)* Whether or not iframes and shadow roots should be traversed when returning the subtree (default is false).
    :returns: Resulting node.
    """
    params: T_JSON_DICT = {}
    if depth is not None:
        params["depth"] = depth
    if pierce is not None:
        params["pierce"] = pierce
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getFlattenedDocument",
        "params": params,
    }
    json = yield cmd_dict
    return [Node.from_json(i) for i in json["nodes"]]


def get_nodes_for_subtree_by_style(
    node_id: NodeId,
    computed_styles: list[CSSComputedStyleProperty],
    pierce: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[NodeId]]:
    """
    Finds nodes with a given computed style in a subtree.

    **EXPERIMENTAL**

    :param node_id: Node ID pointing to the root of a subtree.
    :param computed_styles: The style to filter nodes by (includes nodes if any of properties matches).
    :param pierce: *(Optional)* Whether or not iframes and shadow roots in the same target should be traversed when returning the results (default is false).
    :returns: Resulting nodes.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["computedStyles"] = [i.to_json() for i in computed_styles]
    if pierce is not None:
        params["pierce"] = pierce
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getNodesForSubtreeByStyle",
        "params": params,
    }
    json = yield cmd_dict
    return [NodeId.from_json(i) for i in json["nodeIds"]]


def get_node_for_location(
    x: int,
    y: int,
    include_user_agent_shadow_dom: bool | None = None,
    ignore_pointer_events_none: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, tuple[BackendNodeId, page.FrameId, NodeId | None]]:
    """
    Returns node id at given location. Depending on whether DOM domain is enabled, nodeId is
    either returned or not.

    :param x: X coordinate.
    :param y: Y coordinate.
    :param include_user_agent_shadow_dom: *(Optional)* False to skip to the nearest non-UA shadow root ancestor (default: false).
    :param ignore_pointer_events_none: *(Optional)* Whether to ignore pointer-events: none on elements and hit test them.
    :returns: A tuple with the following items:

        0. **backendNodeId** - Resulting node.
        1. **frameId** - Frame this node belongs to.
        2. **nodeId** - *(Optional)* Id of the node at given coordinates, only when enabled and requested document.
    """
    params: T_JSON_DICT = {}
    params["x"] = x
    params["y"] = y
    if include_user_agent_shadow_dom is not None:
        params["includeUserAgentShadowDOM"] = include_user_agent_shadow_dom
    if ignore_pointer_events_none is not None:
        params["ignorePointerEventsNone"] = ignore_pointer_events_none
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getNodeForLocation",
        "params": params,
    }
    json = yield cmd_dict
    return (
        BackendNodeId.from_json(json["backendNodeId"]),
        page.FrameId.from_json(json["frameId"]),
        NodeId.from_json(json["nodeId"]) if "nodeId" in json else None,
    )


def get_outer_html(
    node_id: NodeId | None = None,
    backend_node_id: BackendNodeId | None = None,
    object_id: runtime.RemoteObjectId | None = None,
    include_shadow_dom: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, str]:
    """
    Returns node's HTML markup.

    :param node_id: *(Optional)* Identifier of the node.
    :param backend_node_id: *(Optional)* Identifier of the backend node.
    :param object_id: *(Optional)* JavaScript object id of the node wrapper.
    :param include_shadow_dom: **(EXPERIMENTAL)** *(Optional)* Include all shadow roots. Equals to false if not specified.
    :returns: Outer HTML markup.
    """
    params: T_JSON_DICT = {}
    if node_id is not None:
        params["nodeId"] = node_id.to_json()
    if backend_node_id is not None:
        params["backendNodeId"] = backend_node_id.to_json()
    if object_id is not None:
        params["objectId"] = object_id.to_json()
    if include_shadow_dom is not None:
        params["includeShadowDOM"] = include_shadow_dom
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getOuterHTML",
        "params": params,
    }
    json = yield cmd_dict
    return str(json["outerHTML"])


def get_relayout_boundary(
    node_id: NodeId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, NodeId]:
    """
    Returns the id of the nearest ancestor that is a relayout boundary.

    **EXPERIMENTAL**

    :param node_id: Id of the node.
    :returns: Relayout boundary node id for the given node.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getRelayoutBoundary",
        "params": params,
    }
    json = yield cmd_dict
    return NodeId.from_json(json["nodeId"])


def get_search_results(
    search_id: str,
    from_index: int,
    to_index: int,
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[NodeId]]:
    """
    Returns search results from given ``fromIndex`` to given ``toIndex`` from the search with the given
    identifier.

    **EXPERIMENTAL**

    :param search_id: Unique search session identifier.
    :param from_index: Start index of the search result to be returned.
    :param to_index: End index of the search result to be returned.
    :returns: Ids of the search result nodes.
    """
    params: T_JSON_DICT = {}
    params["searchId"] = search_id
    params["fromIndex"] = from_index
    params["toIndex"] = to_index
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getSearchResults",
        "params": params,
    }
    json = yield cmd_dict
    return [NodeId.from_json(i) for i in json["nodeIds"]]


def hide_highlight() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Hides any highlight.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.hideHighlight",
    }
    yield cmd_dict


def highlight_node() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Highlights DOM node.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.highlightNode",
    }
    yield cmd_dict


def highlight_rect() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Highlights given rectangle.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.highlightRect",
    }
    yield cmd_dict


def mark_undoable_state() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Marks last undoable state.

    **EXPERIMENTAL**
    """
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.markUndoableState",
    }
    yield cmd_dict


def move_to(
    node_id: NodeId,
    target_node_id: NodeId,
    insert_before_node_id: NodeId | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, NodeId]:
    """
    Moves node into the new container, places it before the given anchor.

    :param node_id: Id of the node to move.
    :param target_node_id: Id of the element to drop the moved node into.
    :param insert_before_node_id: *(Optional)* Drop node before this one (if absent, the moved node becomes the last child of ```targetNodeId```).
    :returns: New id of the moved node.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["targetNodeId"] = target_node_id.to_json()
    if insert_before_node_id is not None:
        params["insertBeforeNodeId"] = insert_before_node_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.moveTo",
        "params": params,
    }
    json = yield cmd_dict
    return NodeId.from_json(json["nodeId"])


def perform_search(
    query: str,
    include_user_agent_shadow_dom: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, tuple[str, int]]:
    """
    Searches for a given string in the DOM tree. Use ``getSearchResults`` to access search results or
    ``cancelSearch`` to end this search session.

    **EXPERIMENTAL**

    :param query: Plain text or query selector or XPath search query.
    :param include_user_agent_shadow_dom: *(Optional)* True to search in user agent shadow DOM.
    :returns: A tuple with the following items:

        0. **searchId** - Unique search session identifier.
        1. **resultCount** - Number of search results.
    """
    params: T_JSON_DICT = {}
    params["query"] = query
    if include_user_agent_shadow_dom is not None:
        params["includeUserAgentShadowDOM"] = include_user_agent_shadow_dom
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.performSearch",
        "params": params,
    }
    json = yield cmd_dict
    return (
        str(json["searchId"]),
        int(json["resultCount"]),
    )


def push_node_by_path_to_frontend(
    path: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, NodeId]:
    """
    Requests that the node is sent to the caller given its path. // FIXME, use XPath

    **EXPERIMENTAL**

    :param path: Path to node in the proprietary format.
    :returns: Id of the node for given path.
    """
    params: T_JSON_DICT = {}
    params["path"] = path
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.pushNodeByPathToFrontend",
        "params": params,
    }
    json = yield cmd_dict
    return NodeId.from_json(json["nodeId"])


def push_nodes_by_backend_ids_to_frontend(
    backend_node_ids: list[BackendNodeId],
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[NodeId]]:
    """
    Requests that a batch of nodes is sent to the caller given their backend node ids.

    **EXPERIMENTAL**

    :param backend_node_ids: The array of backend node ids.
    :returns: The array of ids of pushed nodes that correspond to the backend ids specified in backendNodeIds.
    """
    params: T_JSON_DICT = {}
    params["backendNodeIds"] = [i.to_json() for i in backend_node_ids]
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.pushNodesByBackendIdsToFrontend",
        "params": params,
    }
    json = yield cmd_dict
    return [NodeId.from_json(i) for i in json["nodeIds"]]


def query_selector(
    node_id: NodeId,
    selector: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, NodeId]:
    """
    Executes ``querySelector`` on a given node.

    :param node_id: Id of the node to query upon.
    :param selector: Selector string.
    :returns: Query selector result.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["selector"] = selector
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.querySelector",
        "params": params,
    }
    json = yield cmd_dict
    return NodeId.from_json(json["nodeId"])


def query_selector_all(
    node_id: NodeId,
    selector: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[NodeId]]:
    """
    Executes ``querySelectorAll`` on a given node.

    :param node_id: Id of the node to query upon.
    :param selector: Selector string.
    :returns: Query selector result.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["selector"] = selector
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.querySelectorAll",
        "params": params,
    }
    json = yield cmd_dict
    return [NodeId.from_json(i) for i in json["nodeIds"]]


def get_top_layer_elements() -> Generator[T_JSON_DICT, T_JSON_DICT, list[NodeId]]:
    """
    Returns NodeIds of current top layer elements.
    Top layer is rendered closest to the user within a viewport, therefore its elements always
    appear on top of all other content.

    **EXPERIMENTAL**

    :returns: NodeIds of top layer elements
    """
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getTopLayerElements",
    }
    json = yield cmd_dict
    return [NodeId.from_json(i) for i in json["nodeIds"]]


def get_element_by_relation(
    node_id: NodeId,
    relation: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, NodeId]:
    """
    Returns the NodeId of the matched element according to certain relations.

    **EXPERIMENTAL**

    :param node_id: Id of the node from which to query the relation.
    :param relation: Type of relation to get.
    :returns: NodeId of the element matching the queried relation.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["relation"] = relation
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getElementByRelation",
        "params": params,
    }
    json = yield cmd_dict
    return NodeId.from_json(json["nodeId"])


def redo() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Re-does the last undone action.

    **EXPERIMENTAL**
    """
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.redo",
    }
    yield cmd_dict


def remove_attribute(
    node_id: NodeId,
    name: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Removes attribute with given name from an element with given id.

    :param node_id: Id of the element to remove attribute from.
    :param name: Name of the attribute to remove.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["name"] = name
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.removeAttribute",
        "params": params,
    }
    yield cmd_dict


def remove_node(
    node_id: NodeId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Removes node with given id.

    :param node_id: Id of the node to remove.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.removeNode",
        "params": params,
    }
    yield cmd_dict


def request_child_nodes(
    node_id: NodeId,
    depth: int | None = None,
    pierce: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Requests that children of the node with given id are returned to the caller in form of
    ``setChildNodes`` events where not only immediate children are retrieved, but all children down to
    the specified depth.

    :param node_id: Id of the node to get children for.
    :param depth: *(Optional)* The maximum depth at which children should be retrieved, defaults to 1. Use -1 for the entire subtree or provide an integer larger than 0.
    :param pierce: *(Optional)* Whether or not iframes and shadow roots should be traversed when returning the sub-tree (default is false).
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    if depth is not None:
        params["depth"] = depth
    if pierce is not None:
        params["pierce"] = pierce
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.requestChildNodes",
        "params": params,
    }
    yield cmd_dict


def request_node(
    object_id: runtime.RemoteObjectId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, NodeId]:
    """
    Requests that the node is sent to the caller given the JavaScript node object reference. All
    nodes that form the path from the node to the root are also sent to the client as a series of
    ``setChildNodes`` notifications.

    :param object_id: JavaScript object id to convert into node.
    :returns: Node id for given object.
    """
    params: T_JSON_DICT = {}
    params["objectId"] = object_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.requestNode",
        "params": params,
    }
    json = yield cmd_dict
    return NodeId.from_json(json["nodeId"])


def resolve_node(
    node_id: NodeId | None = None,
    backend_node_id: BackendNodeId | None = None,
    object_group: str | None = None,
    execution_context_id: runtime.ExecutionContextId | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, runtime.RemoteObject]:
    """
    Resolves the JavaScript node object for a given NodeId or BackendNodeId.

    :param node_id: *(Optional)* Id of the node to resolve.
    :param backend_node_id: *(Optional)* Backend identifier of the node to resolve.
    :param object_group: *(Optional)* Symbolic group name that can be used to release multiple objects.
    :param execution_context_id: *(Optional)* Execution context in which to resolve the node.
    :returns: JavaScript object wrapper for given node.
    """
    params: T_JSON_DICT = {}
    if node_id is not None:
        params["nodeId"] = node_id.to_json()
    if backend_node_id is not None:
        params["backendNodeId"] = backend_node_id.to_json()
    if object_group is not None:
        params["objectGroup"] = object_group
    if execution_context_id is not None:
        params["executionContextId"] = execution_context_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.resolveNode",
        "params": params,
    }
    json = yield cmd_dict
    return runtime.RemoteObject.from_json(json["object"])


def set_attribute_value(
    node_id: NodeId,
    name: str,
    value: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Sets attribute for an element with given id.

    :param node_id: Id of the element to set attribute for.
    :param name: Attribute name.
    :param value: Attribute value.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["name"] = name
    params["value"] = value
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.setAttributeValue",
        "params": params,
    }
    yield cmd_dict


def set_attributes_as_text(
    node_id: NodeId,
    text: str,
    name: str | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Sets attributes on element with given id. This method is useful when user edits some existing
    attribute value and types in several attribute name/value pairs.

    :param node_id: Id of the element to set attributes for.
    :param text: Text with a number of attributes. Will parse this text using HTML parser.
    :param name: *(Optional)* Attribute name to replace with new attributes derived from text in case text parsed successfully.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["text"] = text
    if name is not None:
        params["name"] = name
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.setAttributesAsText",
        "params": params,
    }
    yield cmd_dict


def set_file_input_files(
    files: list[str],
    node_id: NodeId | None = None,
    backend_node_id: BackendNodeId | None = None,
    object_id: runtime.RemoteObjectId | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Sets files for the given file input element.

    :param files: Array of file paths to set.
    :param node_id: *(Optional)* Identifier of the node.
    :param backend_node_id: *(Optional)* Identifier of the backend node.
    :param object_id: *(Optional)* JavaScript object id of the node wrapper.
    """
    params: T_JSON_DICT = {}
    params["files"] = list(files)
    if node_id is not None:
        params["nodeId"] = node_id.to_json()
    if backend_node_id is not None:
        params["backendNodeId"] = backend_node_id.to_json()
    if object_id is not None:
        params["objectId"] = object_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.setFileInputFiles",
        "params": params,
    }
    yield cmd_dict


def set_node_stack_traces_enabled(
    enable: bool,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Sets if stack traces should be captured for Nodes. See ``Node.getNodeStackTraces``. Default is disabled.

    **EXPERIMENTAL**

    :param enable: Enable or disable.
    """
    params: T_JSON_DICT = {}
    params["enable"] = enable
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.setNodeStackTracesEnabled",
        "params": params,
    }
    yield cmd_dict


def get_node_stack_traces(
    node_id: NodeId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, runtime.StackTrace | None]:
    """
    Gets stack traces associated with a Node. As of now, only provides stack trace for Node creation.

    **EXPERIMENTAL**

    :param node_id: Id of the node to get stack traces for.
    :returns: *(Optional)* Creation stack trace, if available.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getNodeStackTraces",
        "params": params,
    }
    json = yield cmd_dict
    return runtime.StackTrace.from_json(json["creation"]) if "creation" in json else None


def get_file_info(
    object_id: runtime.RemoteObjectId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, str]:
    """
    Returns file information for the given
    File wrapper.

    **EXPERIMENTAL**

    :param object_id: JavaScript object id of the node wrapper.
    :returns:
    """
    params: T_JSON_DICT = {}
    params["objectId"] = object_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getFileInfo",
        "params": params,
    }
    json = yield cmd_dict
    return str(json["path"])


def get_detached_dom_nodes() -> Generator[T_JSON_DICT, T_JSON_DICT, list[DetachedElementInfo]]:
    """
    Returns list of detached nodes

    **EXPERIMENTAL**

    :returns: The list of detached nodes
    """
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getDetachedDomNodes",
    }
    json = yield cmd_dict
    return [DetachedElementInfo.from_json(i) for i in json["detachedNodes"]]


def set_inspected_node(
    node_id: NodeId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enables console to refer to the node with given id via $x (see Command Line API for more details
    $x functions).

    **EXPERIMENTAL**

    :param node_id: DOM node id to be accessible by means of $x command line API.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.setInspectedNode",
        "params": params,
    }
    yield cmd_dict


def set_node_name(
    node_id: NodeId,
    name: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, NodeId]:
    """
    Sets node name for a node with given id.

    :param node_id: Id of the node to set name for.
    :param name: New node's name.
    :returns: New node's id.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["name"] = name
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.setNodeName",
        "params": params,
    }
    json = yield cmd_dict
    return NodeId.from_json(json["nodeId"])


def set_node_value(
    node_id: NodeId,
    value: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Sets node value for a node with given id.

    :param node_id: Id of the node to set value for.
    :param value: New node's value.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["value"] = value
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.setNodeValue",
        "params": params,
    }
    yield cmd_dict


def set_outer_html(
    node_id: NodeId,
    outer_html: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Sets node HTML markup, returns new node id.

    :param node_id: Id of the node to set markup for.
    :param outer_html: Outer HTML markup to set.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["outerHTML"] = outer_html
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.setOuterHTML",
        "params": params,
    }
    yield cmd_dict


def undo() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Undoes the last performed action.

    **EXPERIMENTAL**
    """
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.undo",
    }
    yield cmd_dict


def get_frame_owner(
    frame_id: page.FrameId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, tuple[BackendNodeId, NodeId | None]]:
    """
    Returns iframe node that owns iframe with the given domain.

    **EXPERIMENTAL**

    :param frame_id:
    :returns: A tuple with the following items:

        0. **backendNodeId** - Resulting node.
        1. **nodeId** - *(Optional)* Id of the node at given coordinates, only when enabled and requested document.
    """
    params: T_JSON_DICT = {}
    params["frameId"] = frame_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getFrameOwner",
        "params": params,
    }
    json = yield cmd_dict
    return (
        BackendNodeId.from_json(json["backendNodeId"]),
        NodeId.from_json(json["nodeId"]) if "nodeId" in json else None,
    )


def get_container_for_node(
    node_id: NodeId,
    container_name: str | None = None,
    physical_axes: PhysicalAxes | None = None,
    logical_axes: LogicalAxes | None = None,
    queries_scroll_state: bool | None = None,
    queries_anchored: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, NodeId | None]:
    """
    Returns the query container of the given node based on container query
    conditions: containerName, physical and logical axes, and whether it queries
    scroll-state or anchored elements. If no axes are provided and
    queriesScrollState is false, the style container is returned, which is the
    direct parent or the closest element with a matching container-name.

    **EXPERIMENTAL**

    :param node_id:
    :param container_name: *(Optional)*
    :param physical_axes: *(Optional)*
    :param logical_axes: *(Optional)*
    :param queries_scroll_state: *(Optional)*
    :param queries_anchored: *(Optional)*
    :returns: *(Optional)* The container node for the given node, or null if not found.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    if container_name is not None:
        params["containerName"] = container_name
    if physical_axes is not None:
        params["physicalAxes"] = physical_axes.to_json()
    if logical_axes is not None:
        params["logicalAxes"] = logical_axes.to_json()
    if queries_scroll_state is not None:
        params["queriesScrollState"] = queries_scroll_state
    if queries_anchored is not None:
        params["queriesAnchored"] = queries_anchored
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getContainerForNode",
        "params": params,
    }
    json = yield cmd_dict
    return NodeId.from_json(json["nodeId"]) if "nodeId" in json else None


def get_querying_descendants_for_container(
    node_id: NodeId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[NodeId]]:
    """
    Returns the descendants of a container query container that have
    container queries against this container.

    **EXPERIMENTAL**

    :param node_id: Id of the container node to find querying descendants from.
    :returns: Descendant nodes with container queries against the given container.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getQueryingDescendantsForContainer",
        "params": params,
    }
    json = yield cmd_dict
    return [NodeId.from_json(i) for i in json["nodeIds"]]


def get_anchor_element(
    node_id: NodeId,
    anchor_specifier: str | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, NodeId]:
    """
    Returns the target anchor element of the given anchor query according to
    https://www.w3.org/TR/css-anchor-position-1/#target.

    **EXPERIMENTAL**

    :param node_id: Id of the positioned element from which to find the anchor.
    :param anchor_specifier: *(Optional)* An optional anchor specifier, as defined in https://www.w3.org/TR/css-anchor-position-1/#anchor-specifier. If not provided, it will return the implicit anchor element for the given positioned element.
    :returns: The anchor element of the given anchor query.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    if anchor_specifier is not None:
        params["anchorSpecifier"] = anchor_specifier
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.getAnchorElement",
        "params": params,
    }
    json = yield cmd_dict
    return NodeId.from_json(json["nodeId"])


def force_show_popover(
    node_id: NodeId,
    enable: bool,
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[NodeId]]:
    """
    When enabling, this API force-opens the popover identified by nodeId
    and keeps it open until disabled.

    **EXPERIMENTAL**

    :param node_id: Id of the popover HTMLElement
    :param enable: If true, opens the popover and keeps it open. If false, closes the popover if it was previously force-opened.
    :returns: List of popovers that were closed in order to respect popover stacking order.
    """
    params: T_JSON_DICT = {}
    params["nodeId"] = node_id.to_json()
    params["enable"] = enable
    cmd_dict: T_JSON_DICT = {
        "method": "DOM.forceShowPopover",
        "params": params,
    }
    json = yield cmd_dict
    return [NodeId.from_json(i) for i in json["nodeIds"]]


@dataclass
class AttributeModified(CDPEvent, event="DOM.attributeModified"):
    """
    Fired when ``Element``'s attribute is modified.
    """
    #: Id of the node that has changed.
    node_id: NodeId
    #: Attribute name.
    name: str
    #: Attribute value.
    value: str

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> AttributeModified:
        return cls(
            node_id=NodeId.from_json(json["nodeId"]),
            name=str(json["name"]),
            value=str(json["value"]),
        )


@dataclass
class AttributeRemoved(CDPEvent, event="DOM.attributeRemoved"):
    """
    Fired when ``Element``'s attribute is removed.
    """
    #: Id of the node that has changed.
    node_id: NodeId
    #: A ttribute name.
    name: str

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> AttributeRemoved:
        return cls(
            node_id=NodeId.from_json(json["nodeId"]),
            name=str(json["name"]),
        )


@dataclass
class CharacterDataModified(CDPEvent, event="DOM.characterDataModified"):
    """
    Mirrors ``DOMCharacterDataModified`` event.
    """
    #: Id of the node that has changed.
    node_id: NodeId
    #: New text value.
    character_data: str

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> CharacterDataModified:
        return cls(
            node_id=NodeId.from_json(json["nodeId"]),
            character_data=str(json["characterData"]),
        )


@dataclass
class ChildNodeCountUpdated(CDPEvent, event="DOM.childNodeCountUpdated"):
    """
    Fired when ``Container``'s child node count has changed.
    """
    #: Id of the node that has changed.
    node_id: NodeId
    #: New node count.
    child_node_count: int

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ChildNodeCountUpdated:
        return cls(
            node_id=NodeId.from_json(json["nodeId"]),
            child_node_count=int(json["childNodeCount"]),
        )


@dataclass
class ChildNodeInserted(CDPEvent, event="DOM.childNodeInserted"):
    """
    Mirrors ``DOMNodeInserted`` event.
    """
    #: Id of the node that has changed.
    parent_node_id: NodeId
    #: Id of the previous sibling.
    previous_node_id: NodeId
    #: Inserted node data.
    node: Node

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ChildNodeInserted:
        return cls(
            parent_node_id=NodeId.from_json(json["parentNodeId"]),
            previous_node_id=NodeId.from_json(json["previousNodeId"]),
            node=Node.from_json(json["node"]),
        )


@dataclass
class ChildNodeRemoved(CDPEvent, event="DOM.childNodeRemoved"):
    """
    Mirrors ``DOMNodeRemoved`` event.
    """
    #: Parent id.
    parent_node_id: NodeId
    #: Id of the node that has been removed.
    node_id: NodeId

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ChildNodeRemoved:
        return cls(
            parent_node_id=NodeId.from_json(json["parentNodeId"]),
            node_id=NodeId.from_json(json["nodeId"]),
        )


@dataclass
class DistributedNodesUpdated(CDPEvent, event="DOM.distributedNodesUpdated"):
    """
    **EXPERIMENTAL**

    Called when distribution is changed.
    """
    #: Insertion point where distributed nodes were updated.
    insertion_point_id: NodeId
    #: Distributed nodes for given insertion point.
    distributed_nodes: list[BackendNode]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> DistributedNodesUpdated:
        return cls(
            insertion_point_id=NodeId.from_json(json["insertionPointId"]),
            distributed_nodes=[BackendNode.from_json(i) for i in json["distributedNodes"]],
        )


@dataclass
class DocumentUpdated(CDPEvent, event="DOM.documentUpdated"):
    """
    Fired when ``Document`` has been totally updated. Node ids are no longer valid.
    """


    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> DocumentUpdated:
        return cls(

        )


@dataclass
class InlineStyleInvalidated(CDPEvent, event="DOM.inlineStyleInvalidated"):
    """
    **EXPERIMENTAL**

    Fired when ``Element``'s inline style is modified via a CSS property modification.
    """
    #: Ids of the nodes for which the inline styles have been invalidated.
    node_ids: list[NodeId]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> InlineStyleInvalidated:
        return cls(
            node_ids=[NodeId.from_json(i) for i in json["nodeIds"]],
        )


@dataclass
class PseudoElementAdded(CDPEvent, event="DOM.pseudoElementAdded"):
    """
    **EXPERIMENTAL**

    Called when a pseudo element is added to an element.
    """
    #: Pseudo element's parent element id.
    parent_id: NodeId
    #: The added pseudo element.
    pseudo_element: Node

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> PseudoElementAdded:
        return cls(
            parent_id=NodeId.from_json(json["parentId"]),
            pseudo_element=Node.from_json(json["pseudoElement"]),
        )


@dataclass
class TopLayerElementsUpdated(CDPEvent, event="DOM.topLayerElementsUpdated"):
    """
    **EXPERIMENTAL**

    Called when top layer elements are changed.
    """


    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> TopLayerElementsUpdated:
        return cls(

        )


@dataclass
class ScrollableFlagUpdated(CDPEvent, event="DOM.scrollableFlagUpdated"):
    """
    **EXPERIMENTAL**

    Fired when a node's scrollability state changes.
    """
    #: The id of the node.
    node_id: NodeId
    #: If the node is scrollable.
    is_scrollable: bool

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ScrollableFlagUpdated:
        return cls(
            node_id=NodeId.from_json(json["nodeId"]),
            is_scrollable=bool(json["isScrollable"]),
        )


@dataclass
class PseudoElementRemoved(CDPEvent, event="DOM.pseudoElementRemoved"):
    """
    **EXPERIMENTAL**

    Called when a pseudo element is removed from an element.
    """
    #: Pseudo element's parent element id.
    parent_id: NodeId
    #: The removed pseudo element id.
    pseudo_element_id: NodeId

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> PseudoElementRemoved:
        return cls(
            parent_id=NodeId.from_json(json["parentId"]),
            pseudo_element_id=NodeId.from_json(json["pseudoElementId"]),
        )


@dataclass
class SetChildNodes(CDPEvent, event="DOM.setChildNodes"):
    """
    Fired when backend wants to provide client with the missing DOM structure. This happens upon
    most of the calls requesting node ids.
    """
    #: Parent node id to populate with children.
    parent_id: NodeId
    #: Child nodes array.
    nodes: list[Node]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> SetChildNodes:
        return cls(
            parent_id=NodeId.from_json(json["parentId"]),
            nodes=[Node.from_json(i) for i in json["nodes"]],
        )


@dataclass
class ShadowRootPopped(CDPEvent, event="DOM.shadowRootPopped"):
    """
    **EXPERIMENTAL**

    Called when shadow root is popped from the element.
    """
    #: Host element id.
    host_id: NodeId
    #: Shadow root id.
    root_id: NodeId

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ShadowRootPopped:
        return cls(
            host_id=NodeId.from_json(json["hostId"]),
            root_id=NodeId.from_json(json["rootId"]),
        )


@dataclass
class ShadowRootPushed(CDPEvent, event="DOM.shadowRootPushed"):
    """
    **EXPERIMENTAL**

    Called when shadow root is pushed into the element.
    """
    #: Host element id.
    host_id: NodeId
    #: Shadow root.
    root: Node

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ShadowRootPushed:
        return cls(
            host_id=NodeId.from_json(json["hostId"]),
            root=Node.from_json(json["root"]),
        )



"""
$description Global live-streaming and video hosting social platform owned by Kick Streaming Pty Ltd.
$url kick.com
$type live, vod
$webbrowser Required for solving a JS challenge that allows access to the Kick API
$metadata id
$metadata author
$metadata category
$metadata title
"""

from __future__ import annotations

import re
from dataclasses import dataclass, replace as dataclass_replace
from ssl import OP_NO_TICKET
from typing import ClassVar

from streamlink.exceptions import NoStreamsError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginargument, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.session.http import SSLContextAdapter
from streamlink.stream.hls import (
    M3U8,
    HLSPlaylist,
    HLSSegment,
    HLSStream,
    HLSStreamReader,
    HLSStreamWorker,
    HLSStreamWriter,
    M3U8Parser,
    parse_tag,
)


log = getLogger(__name__)

LOW_LATENCY_MAX_LIVE_EDGE = 2


@dataclass(kw_only=True)
class KickHLSSegment(HLSSegment):
    prefetch: bool = False


class KickM3U8(M3U8[KickHLSSegment, HLSPlaylist]):
    pass


class KickM3U8Parser(M3U8Parser[KickM3U8, KickHLSSegment, HLSPlaylist]):
    __m3u8__: ClassVar[type[KickM3U8]] = KickM3U8
    __segment__: ClassVar[type[KickHLSSegment]] = KickHLSSegment

    @parse_tag("EXT-X-PREFETCH")
    def parse_tag_ext_x_prefetch(self, value):
        segments = self.m3u8.segments
        if not segments:  # pragma: no cover
            return
        last = segments[-1]

        # Use the average duration of all regular segments for the duration of prefetch segments.
        # This is better than using the duration of the last segment when regular segment durations vary a lot.
        # In low latency mode, the playlist reload time is the duration of the last segment.
        duration = last.duration if last.prefetch else sum(segment.duration for segment in segments) / float(len(segments))

        segment = dataclass_replace(
            last,
            uri=self.uri(value),
            duration=duration,
            title=None,
            prefetch=True,
        )
        segments.append(segment)


class KickHLSStreamWorker(HLSStreamWorker):
    reader: KickHLSStreamReader
    writer: KickHLSStreamWriter
    stream: KickHLSStream

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.stream.low_latency:
            self.reload_time = "segment"

    def process_segments(self, playlist: KickM3U8):  # type: ignore[override, ty:invalid-method-override]
        # ignore prefetch segments if not LL streaming
        if not self.stream.low_latency:
            playlist.segments = [segment for segment in playlist.segments if not segment.prefetch]

        return super().process_segments(playlist)


class KickHLSStreamWriter(HLSStreamWriter):
    reader: KickHLSStreamReader
    stream: KickHLSStream


class KickHLSStreamReader(HLSStreamReader):
    __worker__ = KickHLSStreamWorker
    __writer__ = KickHLSStreamWriter

    worker: KickHLSStreamWorker
    writer: KickHLSStreamWriter
    stream: KickHLSStream

    def __init__(self, stream: KickHLSStream, **kwargs):
        if stream.low_latency:
            live_edge = max(1, min(LOW_LATENCY_MAX_LIVE_EDGE, stream.session.options.get("hls-live-edge")))
            stream.session.options.set("hls-live-edge", live_edge)
            stream.session.options.set("hls-segment-stream-data", True)
            log.info(f"Low latency streaming (HLS live edge: {live_edge})")

        super().__init__(stream, **kwargs)


class KickHLSStream(HLSStream):
    __reader__ = KickHLSStreamReader
    __parser__ = KickM3U8Parser

    def __init__(self, *args, low_latency: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.low_latency = low_latency


class KickAdapter(SSLContextAdapter):
    def get_ssl_context(self):
        ctx = super().get_ssl_context()
        ctx.options &= ~OP_NO_TICKET

        return ctx


@pluginmatcher(
    name="live",
    pattern=re.compile(r"https?://(?:\w+\.)?kick\.com/(?!video/)(?P<channel>[^/?]+)$"),
)
@pluginmatcher(
    name="vod",
    pattern=re.compile(r"https?://(?:\w+\.)?kick\.com/(?:video/|[^/]+/videos/)(?P<vod>[^/?]+)"),
)
@pluginmatcher(
    name="clip",
    pattern=re.compile(r"https?://(?:\w+\.)?kick\.com/(?!video/)(?P<channel>[^/?]+)(?:\?clip=|/clips/)(?P<clip>[^?&]+)"),
)
@pluginargument(
    "low-latency",
    action="store_true",
    help="""
        Enables low latency streaming by prefetching HLS segments.
        Sets --hls-segment-stream-data to true and --hls-live-edge to 2, if it is higher.
        Reducing --hls-live-edge to `1` will result in the lowest latency possible, but will most likely cause buffering.

        In order to achieve true low latency streaming during playback, the player's caching/buffering settings will
        need to be adjusted and reduced to a value as low as possible, but still high enough to not cause any buffering.
        This depends on the stream's bitrate and the quality of the connection to Kick's servers. Please refer to the
        player's own documentation for the required configuration. Player parameters can be set via --player-args.

        Note: Low latency streams have to be enabled by the broadcasters on Kick themselves.
        Regular streams can cause buffering issues with this option enabled due to the reduced --hls-live-edge value.
    """,
)
class Kick(Plugin):
    _CACHE_HEADERS = "headers"
    _CACHE_EXPIRATION = 3600 * 24 * 30

    _URL_API_LIVESTREAM = "https://kick.com/api/v2/channels/{channel}/livestream"
    _URL_API_VOD = "https://kick.com/api/v1/video/{vod}"
    _URL_API_CLIP = "https://kick.com/api/v2/clips/{clip}"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session.http.mount("https://kick.com/", KickAdapter())
        self.session.http.headers.update(self.cache.get(self._CACHE_HEADERS, {}))
        self.session.http.headers.update({
            "Referer": self.url,
        })

    def _get_cookies_from_webbrowser(self) -> bool:
        from streamlink.compat import BaseExceptionGroup  # noqa: PLC0415
        from streamlink.webbrowser.cdp import CDPClient, CDPClientSession  # noqa: PLC0415, TC001
        from streamlink.webbrowser.cdp.devtools import fetch  # noqa: PLC0415, TC001

        async def on_main(client_session: CDPClientSession, request: fetch.RequestPaused):
            # get Chromium's request headers, update HTTP session headers and also cache them
            self.session.http.headers.update(request.request.headers)
            self.cache.set(self._CACHE_HEADERS, request.request.headers, expires=self._CACHE_EXPIRATION)
            await client_session.continue_request(request)

        async def get_challenge_cookies(client: CDPClient):
            client_session: CDPClientSession
            async with client.session() as client_session:
                client_session.add_request_handler(on_main, url_pattern=self.url, on_request=True)
                async with client_session.navigate(self.url) as frame_id:
                    await client_session.loaded(frame_id)
                    await client_session.retrieve_cookies()

        log.info("Solving JS challenge")

        try:
            CDPClient.launch(self.session, get_challenge_cookies)
        except BaseExceptionGroup:
            log.exception("Failed solving JS challenge")
        except Exception as err:
            log.error(err)
        else:
            log.info("JS challenge solved, storing cookies")
            self.save_cookies()

            return True

        return False

    def _get_api_headers(self):
        # _get_cookies_from_webbrowser() from above updates the session headers with Chromium's initial request headers
        ua = str(self.session.http.headers.get("User-Agent", useragents.CHROME))
        if "Chrome/" not in ua:
            ua = useragents.CHROME

        if not (m := re.search(r"Chrome/(?P<full>(?P<main>\d+)\S+)", ua)):
            raise PluginError("Error while parsing Chromium User-Agent")

        return {
            "Accept": "application/json",
            "Accept-Language": "en-US",
            "sec-ch-ua": f'"Not:A-Brand";v="24", "Chromium";v="{m["main"]}"',
            "sec-ch-ua-arch": '"x86"',
            "sec-ch-ua-bitness": '"64"',
            "sec-ch-ua-full-version": f'"{m["full"]}"',
            "sec-ch-ua-full-version-list": f'"Not:A-Brand";v="24.0.0.0", "Chromium";v="{m["full"]}"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-model": '""',
            "sec-ch-ua-platform": '"Windows"',
            "sec-ch-ua-platform-version": '"6.14.0"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "sec-fetch-user": "?1",
        }

    def _query_api(self, url, schema, secondary_attempt=False):
        res = self.session.http.get(
            url,
            headers=self._get_api_headers(),
            raise_for_status=False,
        )
        if res.status_code == 403 and not secondary_attempt and self._get_cookies_from_webbrowser():
            # re-attempt API query after getting (new) cookies
            return self._query_api(url, schema, secondary_attempt=True)

        try:
            res.raise_for_status()
        except Exception as err:
            raise PluginError(f"Error while querying Kick API: {err or '403 status response'}") from err

        main_schema = validate.Schema(
            validate.parse_json(),
            validate.any(
                validate.all(
                    {"message": str},
                    validate.transform(lambda obj: ("error", obj["message"])),
                ),
                validate.all(
                    {"data": None},
                    validate.transform(lambda _: ("data", None)),
                ),
                validate.all(
                    schema,
                    validate.transform(lambda obj: ("data", obj)),
                ),
            ),
        )
        restype, data = main_schema.validate(res.text)

        if restype == "error":
            raise PluginError(f"Error while querying Kick API: {data or 'unknown error'}")
        if not data:
            raise NoStreamsError

        return data

    def _get_streams_live(self):
        self.author = self.match["channel"]

        hls_url, self.id, self.category, self.title = self._query_api(
            self._URL_API_LIVESTREAM.format(channel=self.author),
            schema=validate.Schema(
                {
                    "data": {
                        "playback_url": validate.url(path=validate.endswith(".m3u8")),
                        "id": int,
                        "category": {"name": str},
                        "session_title": str,
                    },
                },
                validate.get("data"),
                validate.union_get(
                    "playback_url",
                    "id",
                    ("category", "name"),
                    "session_title",
                ),
            ),
        )

        return KickHLSStream.parse_variant_playlist(self.session, hls_url, low_latency=self.get_option("low-latency"))

    def _get_streams_vod(self):
        self.id = self.match["vod"]

        hls_url, self.author, self.title = self._query_api(
            self._URL_API_VOD.format(vod=self.id),
            schema=validate.Schema(
                {
                    "source": validate.url(path=validate.endswith(".m3u8")),
                    "livestream": {
                        "session_title": str,
                        "channel": {
                            "user": {
                                "username": str,
                            },
                        },
                    },
                },
                validate.union_get(
                    "source",
                    ("livestream", "channel", "user", "username"),
                    ("livestream", "session_title"),
                ),
            ),
        )

        return HLSStream.parse_variant_playlist(self.session, hls_url)

    def _get_streams_clip(self):
        self.id = self.match["clip"]
        self.author = self.match["channel"]

        hls_url, self.category, self.title = self._query_api(
            self._URL_API_CLIP.format(clip=self.id),
            schema=validate.Schema(
                {
                    "clip": {
                        "clip_url": validate.url(path=validate.endswith(".m3u8")),
                        "category": {"name": str},
                        "title": str,
                    },
                },
                validate.get("clip"),
                validate.union_get(
                    "clip_url",
                    ("category", "name"),
                    "title",
                ),
            ),
        )

        return {"clip": HLSStream(self.session, hls_url)}

    def _get_streams(self):
        if self.matches["live"]:
            return self._get_streams_live()
        if self.matches["vod"]:
            return self._get_streams_vod()
        if self.matches["clip"]:
            return self._get_streams_clip()


__plugin__ = Kick



"""
$description Russian live streaming platform hosting various Russian live TV channels.
$url mediavitrina.ru
$type live
$region Russia
"""

import re
from urllib.parse import urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_qsd


log = getLogger(__name__)


@pluginmatcher(
    name="chetv",
    pattern=re.compile(r"https?://(?:www\.)?chetv\.ru/(?:live|online)"),
)
@pluginmatcher(
    name="ctc",
    pattern=re.compile(r"https?://(?:www\.)?ctc(?:love)?\.ru/(?:live|online)"),
)
@pluginmatcher(
    name="domashniy",
    pattern=re.compile(r"https?://(?:www\.)?domashniy\.ru/(?:live|online)"),
)
@pluginmatcher(
    name="mediavitrina",
    pattern=re.compile(r"https?://player\.mediavitrina\.ru/.+/player\.html"),
)
class MediaVitrina(Plugin):
    _re_url_json = re.compile(r"https://media\.mediavitrina\.ru/(?:proxy)?api/v3/[\w-]+/playlist/[\w-]+_as_array\.json[^\"']+")

    def _get_streams(self):
        self.session.http.headers.update({"Referer": self.url})

        p_netloc = urlparse(self.url).netloc
        if p_netloc == "player.mediavitrina.ru":
            # https://player.mediavitrina.ru/
            url_player = self.url
        elif p_netloc.endswith("ctc.ru"):
            # https://ctc.ru/online/
            url_player = self.session.http.get(
                "https://ctc.ru/api/page/v1/online/",
                schema=validate.Schema(
                    validate.parse_json(),
                    {
                        "content": validate.all(
                            [dict],
                            validate.filter(lambda n: n.get("type") == "on-air"),
                            [{"onAirLink": validate.url(netloc="player.mediavitrina.ru")}],
                            validate.get((0, "onAirLink")),
                        ),
                    },
                    validate.get("content"),
                ),
            )
        else:
            # https://chetv.ru/online/
            # https://ctclove.ru/online/
            # https://domashniy.ru/online/
            url_player = self.session.http.get(
                self.url,
                schema=validate.Schema(
                    validate.parse_html(),
                    validate.xml_xpath_string(".//iframe[starts-with(@src,'https://player.mediavitrina.ru/')]/@src"),
                ),
                acceptable_status=(200, 403, 404),
            )

        if not url_player:
            return

        log.debug(f"url_player={url_player}")
        script_data = self.session.http.get(
            url_player,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[contains(text(),'media.mediavitrina.ru/')]/text()"),
            ),
        )
        if not script_data:
            log.debug("invalid script_data")
            return

        m = self._re_url_json.search(script_data)
        if not m:
            log.debug("invalid url_json")
            return

        url_json = m.group(0)
        log.debug(f"url_json={url_json}")
        url_json = re.sub(r"\{\{PLAYER_REFERER_HOSTNAME\}\}", "mediavitrina.ru", url_json)
        url_json = re.sub(r"\{\{[A-Za-z_]+\}\}", "", url_json)

        res_token = self.session.http.get(
            "https://media.mediavitrina.ru/get_token",
            schema=validate.Schema(
                validate.parse_json(),
                {"result": {"token": str}},
                validate.get("result"),
            ),
        )
        url = self.session.http.get(
            update_qsd(url_json, qsd=res_token),
            schema=validate.Schema(
                validate.parse_json(),
                {"hls": [validate.url()]},
                validate.get(("hls", 0)),
            ),
        )

        if not url:
            return

        if "georestrictions" in url:
            log.error("Stream is geo-restricted")
            return

        return HLSStream.parse_variant_playlist(self.session, url, name_fmt="{pixels}_{bitrate}")


__plugin__ = MediaVitrina



from __future__ import annotations

import json
import logging
from threading import Event, RLock, Thread, current_thread
from typing import TYPE_CHECKING, Any, TypedDict
from urllib.parse import unquote_plus, urlparse

from certifi import where as certify_where
from websocket import ABNF, STATUS_NORMAL, WebSocketApp, enableTrace

from streamlink.logger import TRACE, getLogger, root as rootlogger


if TYPE_CHECKING:
    from typing_extensions import NotRequired

    from streamlink.session import Streamlink

    class TWSRunForever(TypedDict):
        sockopt: NotRequired[tuple]
        sslopt: NotRequired[dict]
        ping_interval: NotRequired[float | int]
        ping_timeout: NotRequired[float | int | None]
        ping_payload: NotRequired[str]
        http_proxy_host: NotRequired[str]
        http_proxy_port: NotRequired[int | str]
        http_no_proxy: NotRequired[list]
        http_proxy_auth: NotRequired[tuple]
        http_proxy_timeout: NotRequired[float | None]
        skip_utf8_validation: NotRequired[bool]
        host: NotRequired[str]
        origin: NotRequired[str]
        suppress_origin: NotRequired[bool]
        proxy_type: NotRequired[str]
        reconnect: NotRequired[int]


log = getLogger(__name__)


class WebsocketClient(Thread):
    OPCODE_CONT: int = ABNF.OPCODE_CONT
    OPCODE_TEXT: int = ABNF.OPCODE_TEXT
    OPCODE_BINARY: int = ABNF.OPCODE_BINARY
    OPCODE_CLOSE: int = ABNF.OPCODE_CLOSE
    OPCODE_PING: int = ABNF.OPCODE_PING
    OPCODE_PONG: int = ABNF.OPCODE_PONG

    _id: int = 0

    ws: WebSocketApp

    def __init__(
        self,
        session: Streamlink,
        url: str,
        subprotocols: list[str] | None = None,
        header: list[str] | dict[str, str] | None = None,
        cookie: str | None = None,
        sockopt: tuple | None = None,
        sslopt: dict | None = None,
        host: str | None = None,
        origin: str | None = None,
        suppress_origin: bool = False,
        ping_interval: int | float = 0,
        ping_timeout: int | float | None = None,
        ping_payload: str = "",
    ):
        if rootlogger.level <= TRACE:
            enableTrace(
                True,
                handler=next(
                    (handler for handler in rootlogger.handlers if isinstance(handler, logging.StreamHandler)),
                    logging.StreamHandler(),
                ),
            )

        if not header:
            header = []
        elif isinstance(header, dict):
            header = [f"{k!s}: {v!s}" for k, v in header.items()]
        if not any(True for h in header if h.startswith("User-Agent: ")):
            header.append(f"User-Agent: {session.http.headers['User-Agent']!s}")

        self.reconnect_done = Event()
        self.is_reconnecting = Event()
        self._reconnect_lock = RLock()

        if not sslopt:  # pragma: no cover
            sslopt = {}
        sslopt.setdefault("ca_certs", certify_where())

        self.session = session
        self._ws_init(url, subprotocols, header, cookie)
        self._ws_rundata: TWSRunForever = dict(
            sslopt=sslopt,
            suppress_origin=suppress_origin,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            ping_payload=ping_payload,
        )

        if sockopt:  # pragma: no branch
            self._ws_rundata["sockopt"] = sockopt
        if host:  # pragma: no branch
            self._ws_rundata["host"] = host
        if origin:  # pragma: no branch
            self._ws_rundata["origin"] = origin

        http_proxy: str | None = session.get_option("http-proxy")
        if http_proxy:
            p = urlparse(http_proxy)
            self._ws_rundata["proxy_type"] = p.scheme
            if p.hostname:  # pragma: no branch
                self._ws_rundata["http_proxy_host"] = p.hostname
            if p.port:  # pragma: no branch
                self._ws_rundata["http_proxy_port"] = p.port
            if p.username:  # pragma: no branch
                self._ws_rundata["http_proxy_auth"] = unquote_plus(p.username), unquote_plus(p.password or "")

        self._id += 1
        super().__init__(
            name=f"Thread-{self.__class__.__name__}-{self._id}",
            daemon=True,
        )

    def _ws_init(self, url, subprotocols, header, cookie):
        self.ws = WebSocketApp(
            url=url,
            subprotocols=subprotocols,
            header=header,
            cookie=cookie,
            on_open=self.on_open,
            on_error=self.on_error,
            on_close=self.on_close,
            on_ping=self.on_ping,
            on_pong=self.on_pong,
            on_message=self.on_message,
            on_cont_message=self.on_cont_message,
            on_data=self.on_data,
        )

    def run(self) -> None:
        while True:
            log.debug(f"Connecting to: {self.ws.url}")
            self.ws.run_forever(**self._ws_rundata)
            # check if closed via a reconnect() call
            with self._reconnect_lock:
                if not self.is_reconnecting.is_set():
                    return
                self.reconnect_done.set()
                self.is_reconnecting.clear()

    # ----

    def reconnect(
        self,
        url: str | None = None,
        subprotocols: list[str] | None = None,
        header: list | dict | None = None,
        cookie: str | None = None,
        closeopts: dict | None = None,
    ) -> None:
        with self._reconnect_lock:
            # ws connection is not active (anymore)
            if not self.ws.keep_running:
                return
            if self.is_reconnecting.is_set():
                return
            self.is_reconnecting.set()
            self.reconnect_done.clear()
            log.debug("Reconnecting...")
            self.ws.close(**(closeopts or {}))
            self._ws_init(
                url=self.ws.url if url is None else url,
                subprotocols=self.ws.subprotocols if subprotocols is None else subprotocols,
                header=self.ws.header if header is None else header,
                cookie=self.ws.cookie if cookie is None else cookie,
            )

    def close(self, status: int = STATUS_NORMAL, reason: str | bytes = "", timeout: int = 3) -> None:
        if isinstance(reason, str):
            reason = bytes(reason, encoding="utf-8")
        self.ws.close(status=status, reason=reason, timeout=timeout)
        if self.is_alive() and current_thread() is not self:
            self.join()

    def send(self, data: str | bytes, opcode: int = ABNF.OPCODE_TEXT) -> None:
        return self.ws.send(data, opcode)

    def send_json(self, data: Any) -> None:
        return self.send(json.dumps(data, indent=None, separators=(",", ":")))

    # ----

    # noinspection PyMethodMayBeStatic
    def on_open(self, wsapp: WebSocketApp) -> None:
        log.debug(f"Connected: {wsapp.url}")  # pragma: no cover

    # noinspection PyMethodMayBeStatic
    # noinspection PyUnusedLocal
    def on_error(self, wsapp: WebSocketApp, error: Exception) -> None:
        log.error(error)  # pragma: no cover

    # noinspection PyMethodMayBeStatic
    # noinspection PyUnusedLocal
    def on_close(self, wsapp: WebSocketApp, status: int, message: str) -> None:
        log.debug(f"Closed: {wsapp.url}")  # pragma: no cover

    def on_ping(self, wsapp: WebSocketApp, data: bytes) -> None:
        pass  # pragma: no cover

    def on_pong(self, wsapp: WebSocketApp, data: bytes) -> None:
        pass  # pragma: no cover

    def on_message(self, wsapp: WebSocketApp, data: str) -> None:
        pass  # pragma: no cover

    def on_cont_message(self, wsapp: WebSocketApp, data: bytes, cont: Any) -> None:
        pass  # pragma: no cover

    def on_data(self, wsapp: WebSocketApp, data: bytes | str, data_type: int, cont: Any) -> None:
        pass  # pragma: no cover



"""
$description Live TV channels from MTVA, a Hungarian public, state-owned broadcaster.
$url mediaklikk.hu
$url m4sport.hu
$url hirado.hu
$type live
$region Hungary
"""

import re
from urllib.parse import unquote, urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?(?:mediaklikk|m4sport|hirado)\.hu/"),
)
class Mediaklikk(Plugin):
    PLAYER_URL = "https://player.mediaklikk.hu/playernew/player.php"

    def _get_streams(self):
        params = self.session.http.get(
            self.url,
            schema=validate.Schema(
                re.compile(
                    r"""
                        loadPlayer\s*\(\s*
                            (?P<q>['"]).+?(?P=q)\s*,\s*
                            (?P<json>{.*?})\s*
                        \)
                    """,
                    re.VERBOSE | re.DOTALL,
                ),
                validate.none_or_all(
                    validate.get("json"),
                    validate.parse_json(),
                    {
                        "contentId": validate.any(str, int),
                        validate.optional("streamId"): str,
                        validate.optional("idec"): str,
                        validate.optional("token"): str,
                    },
                ),
            ),
        )
        if not params:
            log.error("Could not find player data")
            return

        params.update({
            "video": (
                unquote(params.pop("token"))
                if params.get("token") is not None
                else params.pop("streamId")
            ),
            "noflash": "yes",
            "embedded": "0",
        })  # fmt: skip

        url_parsed = urlparse(self.url)
        skip_vods = url_parsed.netloc.endswith("m4sport.hu") and url_parsed.path.startswith("/elo")

        self.session.http.headers.update({"Referer": self.url})
        playlists = self.session.http.get(
            self.PLAYER_URL,
            params=params,
            schema=validate.Schema(
                re.compile(r"pl\.setup\s*\(\s*(?P<json>{.*?})\s*\)\s*;", re.DOTALL),
                validate.none_or_all(
                    validate.get("json"),
                    validate.parse_json(),
                    {
                        "playlist": [
                            {
                                "file": validate.url(),
                                "type": str,
                            },
                        ],
                    },
                    validate.get("playlist"),
                    validate.filter(lambda p: p["type"] == "hls"),
                    validate.filter(lambda p: not skip_vods or "vod" not in p["file"]),
                    validate.map(lambda p: update_scheme("https://", p["file"])),
                ),
            ),
        )

        for url in playlists or []:
            yield from HLSStream.parse_variant_playlist(self.session, url).items()


__plugin__ = Mediaklikk



from __future__ import annotations

import copy
import math
import re
from collections import defaultdict
from contextlib import contextmanager
from datetime import timedelta
from itertools import count, repeat
from typing import TYPE_CHECKING, Any, ClassVar, Literal, TypeAlias, TypeVar, overload
from urllib.parse import urljoin, urlparse, urlunparse

from isodate import Duration, parse_datetime, parse_duration  # type: ignore[import]

from streamlink.logger import getLogger
from streamlink.stream.dash.segment import DASHSegment, TimelineSegment
from streamlink.utils.times import UTC, fromtimestamp, now
from streamlink.utils.url import is_insecure_scheme


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Sequence
    from datetime import datetime

    # noinspection PyProtectedMember
    from lxml.etree import _Attrib, _Element

    TMPDNode_co = TypeVar("TMPDNode_co", bound="MPDNode", covariant=True)
    TAttrDefault = TypeVar("TAttrDefault")
    TAttrParseResult = TypeVar("TAttrParseResult")
    TAttrInherited: TypeAlias = type[TMPDNode_co] | Sequence[type[TMPDNode_co]] | None
    TTimelineIdent: TypeAlias = tuple[str | None, str | None, str]


log = getLogger(__name__)

EPOCH_START = fromtimestamp(0)
ONE_SECOND = timedelta(seconds=1)


def _identity(x):
    return x


def datetime_to_seconds(dt):
    return (dt - EPOCH_START).total_seconds()


def count_dt(firstval: datetime | None = None, step: timedelta = ONE_SECOND) -> Iterator[datetime]:
    current = now() if firstval is None else firstval
    while True:
        yield current
        current += step


@contextmanager
def freeze_timeline(mpd):
    timelines = copy.copy(mpd.timelines)
    yield
    mpd.timelines = timelines


_re_segment_template = re.compile(r"(.*?)\$(\w+)(?:%([\w.]+))?\$")


class MPDParsers:
    @staticmethod
    def bool_str(v: str) -> bool:
        return v.lower() == "true"

    @staticmethod
    def type(mpdtype: Literal["static", "dynamic"]) -> Literal["static", "dynamic"]:
        if mpdtype not in ("static", "dynamic"):
            raise MPDParsingError("@type must be static or dynamic")
        return mpdtype

    @staticmethod
    def duration(anchor: datetime | None = None) -> Callable[[str], timedelta]:
        def duration_to_timedelta(duration: str) -> timedelta:
            parsed: timedelta | Duration = parse_duration(duration)
            if isinstance(parsed, Duration):
                return parsed.totimedelta(start=anchor or now())
            return parsed

        return duration_to_timedelta

    @staticmethod
    def parse_datetime(dt: str) -> datetime:
        return parse_datetime(dt).replace(tzinfo=UTC)

    @staticmethod
    def parse_timedelta(timescale: int | float = 1.0) -> Callable[[str | int | float], timedelta]:
        def _timedelta(seconds: str | int | float):
            return timedelta(seconds=int(float(seconds) / float(timescale)))

        return _timedelta

    @staticmethod
    def segment_template(url_template: str) -> Callable[..., str]:
        end = 0
        res = ""
        for m in _re_segment_template.finditer(url_template):
            _, end = m.span()
            res += f"{m[1]}{{{m[2]}{f':{m[3]}' if m[3] else ''}}}"

        return f"{res}{url_template[end:]}".format

    @staticmethod
    def frame_rate(frame_rate: str) -> float:
        if "/" not in frame_rate:
            return float(frame_rate)

        a, b = frame_rate.split("/")
        return float(a) / float(b)

    @staticmethod
    def bandwidth(bandwidth: str) -> float:
        return float(bandwidth) / 1000.0

    @staticmethod
    def range(range_spec: str) -> tuple[int, int | None]:
        r = range_spec.split("-")
        if len(r) != 2:
            raise MPDParsingError("Invalid byte-range-spec")

        start = int(r[0])
        if r[1]:
            return start, int(r[1]) - start + 1
        else:
            return start, None


class MPDParsingError(Exception):
    pass


class MPDNode:
    __tag__: ClassVar[str]

    parent: MPDNode
    _base_url: str
    baseURLs: list[BaseURL]

    def __init__(self, node: _Element, root: MPD, parent: MPDNode, **kwargs) -> None:
        self.node = node
        self.root = root
        self.parent = parent
        self._base_url = kwargs.get("base_url", "")
        self.attributes: set[str] = set()
        if self.__tag__ and self.node.tag.lower() != self.__tag__.lower():
            raise MPDParsingError(f"Root tag did not match the expected tag: {self.__tag__}")

    @property
    def attrib(self) -> _Attrib:
        return self.node.attrib

    @property
    def text(self) -> str | None:
        return self.node.text

    def __str__(self):
        return f"<{self.__tag__} {' '.join(f'@{attr}={getattr(self, attr)}' for attr in self.attributes)}>"

    @overload
    def attr(
        self,
        key: str,
        inherited: TAttrInherited = None,
    ) -> str | None: ...  # pragma: no cover

    @overload
    def attr(
        self,
        key: str,
        parser: type[TAttrParseResult] | Callable[[Any], TAttrParseResult],
        inherited: TAttrInherited = None,
    ) -> TAttrParseResult | None: ...  # pragma: no cover

    @overload
    def attr(
        self,
        key: str,
        required: Literal[True],
        inherited: TAttrInherited = None,
    ) -> str: ...  # pragma: no cover

    @overload
    def attr(
        self,
        key: str,
        parser: type[TAttrParseResult] | Callable[[Any], TAttrParseResult],
        required: Literal[True],
        inherited: TAttrInherited = None,
    ) -> TAttrParseResult: ...  # pragma: no cover

    @overload
    def attr(
        self,
        key: str,
        required: bool = False,
        inherited: TAttrInherited = None,
    ) -> str | None: ...  # pragma: no cover

    @overload
    def attr(
        self,
        key: str,
        default: TAttrDefault,
        inherited: TAttrInherited = None,
    ) -> str | TAttrDefault: ...  # pragma: no cover

    @overload
    def attr(
        self,
        key: str,
        parser: type[TAttrParseResult] | Callable[[Any], TAttrParseResult],
        default: TAttrDefault,
        inherited: TAttrInherited = None,
    ) -> TAttrParseResult | TAttrDefault: ...  # pragma: no cover

    def attr(self, key, parser=None, default=None, required=False, inherited=None):
        self.attributes.add(key)
        if key in self.attrib:
            value = self.attrib.get(key)
            if parser and callable(parser):
                return parser(value)
            else:
                return value
        elif inherited:
            value = self.walk_back_get_attr(key, inherited)
            if value is not None:
                return value

        if required:  # pragma: no cover
            raise MPDParsingError(f"Could not find required attribute {self.__tag__}@{key} ")

        return default

    def children(
        self,
        cls: type[TMPDNode_co],
        minimum: int = 0,
        maximum: int | None = None,
        **kwargs,
    ) -> list[TMPDNode_co]:
        children = self.node.findall(cls.__tag__)
        if len(children) < minimum or (maximum and len(children) > maximum):
            raise MPDParsingError(f"Expected to find {self.__tag__}/{cls.__tag__} required [{minimum}..{maximum or 'unbound'})")

        return [
            cls(child, root=self.root, parent=self, i=i, base_url=self.base_url, **kwargs)
            for i, child in enumerate(children)
        ]  # fmt: skip

    def only_child(
        self,
        cls: type[TMPDNode_co],
        minimum: int = 0,
        **kwargs,
    ) -> TMPDNode_co | None:
        children = self.children(cls, minimum=minimum, maximum=1, **kwargs)
        return children[0] if len(children) else None

    def walk_back(
        self,
        cls: type[TMPDNode_co] | Sequence[type[TMPDNode_co]] | None = None,
        mapper: Callable[[MPDNode], MPDNode | None] = _identity,
    ) -> Iterator[MPDNode]:
        node = self.parent
        while node:
            if cls is None or isinstance(node, cls):  # type: ignore[arg-type, ty:invalid-argument-type]
                n = mapper(node)
                if n is not None:
                    yield n
            node = node.parent

    def walk_back_get_attr(
        self,
        attr: str,
        cls: type[TMPDNode_co] | Sequence[type[TMPDNode_co]] | None = None,
        mapper: Callable[[MPDNode], MPDNode | None] = _identity,
    ) -> Any | None:
        for ancestor in self.walk_back(cls, mapper):
            value = getattr(ancestor, attr, None)
            if value is not None:
                return value

    @property
    def base_url(self) -> str:
        if hasattr(self, "baseURLs") and self.baseURLs and (other := self.baseURLs[0].url):
            base_scheme = urlparse(self._base_url).scheme
            scheme = urlparse(other).scheme
            if is_insecure_scheme(base_scheme, scheme):
                raise MPDParsingError(f"Prevented access to insecure resource in manifest: {base_scheme=!r} {scheme=!r}")

            return urljoin(self._base_url, other)

        return self._base_url


class MPD(MPDNode):
    """
    Represents the MPD as a whole

    Should validate the XML input and provide methods to get segment URLs for each Period, AdaptationSet and Representation.
    """

    __tag__ = "MPD"

    parent: None  # type: ignore[assignment]
    timelines: dict[TTimelineIdent, int]

    DEFAULT_MINBUFFERTIME = 3.0
    DEFAULT_LIVE_EDGE_SEGMENTS = 3

    def __init__(self, *args, url: str | None = None, **kwargs) -> None:
        # top level has no parent
        kwargs["root"] = self
        kwargs["parent"] = None
        super().__init__(*args, **kwargs)

        # parser attributes
        self.url = url
        self.timelines = defaultdict(lambda: -1)
        self.timelines.update(kwargs.pop("timelines", {}))

        self.id = self.attr("id")
        self.profiles = self.attr(
            "profiles",
            required=True,
        )
        self.type = self.attr(
            "type",
            parser=MPDParsers.type,
            default="static",
        )
        self.publishTime = self.attr(
            "publishTime",
            parser=MPDParsers.parse_datetime,
            default=EPOCH_START,
        )
        self.availabilityStartTime = self.attr(
            "availabilityStartTime",
            parser=MPDParsers.parse_datetime,
            default=EPOCH_START,
        )
        self.availabilityEndTime = self.attr(
            "availabilityEndTime",
            parser=MPDParsers.parse_datetime,
        )
        self.minBufferTime = self.attr(
            "minBufferTime",
            parser=MPDParsers.duration(self.publishTime),
            required=True,
        )
        self.minimumUpdatePeriod = self.attr(
            "minimumUpdatePeriod",
            parser=MPDParsers.duration(self.publishTime),
            default=timedelta(),
        )
        self.timeShiftBufferDepth = self.attr(
            "timeShiftBufferDepth",
            parser=MPDParsers.duration(self.publishTime),
        )
        self.mediaPresentationDuration = self.attr(
            "mediaPresentationDuration",
            parser=MPDParsers.duration(self.publishTime),
            default=timedelta(),
        )
        self.suggestedPresentationDelay = self.attr(
            "suggestedPresentationDelay",
            parser=MPDParsers.duration(self.publishTime),
            # if there is no delay, use a delay of 3 seconds, but respect the manifest's minBufferTime
            # TODO: add a customizable parameter for this
            default=timedelta(
                seconds=max(
                    self.DEFAULT_MINBUFFERTIME,
                    self.minBufferTime.total_seconds(),
                ),
            ),
        )

        # parse children
        location = self.children(Location)
        self.location = location[0] if location else None
        if self.location:
            self.url = self.location.text or ""
            urlp = list(urlparse(self.url))
            if urlp[2]:
                urlp[2], _ = urlp[2].rsplit("/", 1)
            self._base_url = urlunparse(urlp)

        self.baseURLs = self.children(BaseURL)
        self.periods = self.children(Period, minimum=1)
        self.periods_map = {period.id: period for period in self.periods if period.id is not None}
        self.programInformation = self.children(ProgramInformation)

    def get_representation(self, ident: TTimelineIdent) -> Representation | None:
        """
        Find the first Representation instance with a matching ident
        """
        p, a, r = ident
        for period in self.periods:
            if p != period.id:
                continue
            for adaptationset in period.adaptationSets:
                if a != adaptationset.id:
                    continue
                for representation in adaptationset.representations:
                    if r != representation.id:
                        continue
                    return representation


class ProgramInformation(MPDNode):
    __tag__ = "ProgramInformation"


class BaseURL(MPDNode):
    __tag__ = "BaseURL"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.url = (self.text or "").strip()


class Location(MPDNode):
    __tag__ = "Location"


class Period(MPDNode):
    __tag__ = "Period"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.i = kwargs.get("i", 0)
        self.id = self.attr("id")
        self.bitstreamSwitching = self.attr(
            "bitstreamSwitching",
            parser=MPDParsers.bool_str,
        )
        self.duration = self.attr(
            "duration",
            parser=MPDParsers.duration(self.root.publishTime),
            default=timedelta(),
        )
        self.start = self.attr(
            "start",
            parser=MPDParsers.duration(self.root.publishTime),
            default=timedelta(),
        )

        # anchor time for segment availability
        offset = self.start if self.root.type == "dynamic" else timedelta()
        self.availabilityStartTime = self.root.availabilityStartTime + offset

        # TODO: Early Access Periods

        self.baseURLs = self.children(BaseURL)
        self.segmentBase = self.only_child(SegmentBase, period=self)
        self.segmentList = self.only_child(SegmentList, period=self)
        self.segmentTemplate = self.only_child(SegmentTemplate, period=self)
        self.adaptationSets = self.children(AdaptationSet, minimum=1)
        self.assetIdentifier = self.only_child(AssetIdentifier)
        self.eventStream = self.children(EventStream)
        self.subset = self.children(Subset)


class AssetIdentifier(MPDNode):
    __tag__ = "AssetIdentifier"


class Subset(MPDNode):
    __tag__ = "Subset"


class EventStream(MPDNode):
    __tag__ = "EventStream"


class _RepresentationBaseType(MPDNode):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # mimeType must be set on the AdaptationSet or Representation
        self.mimeType = self.attr(
            "mimeType",
            required=type(self) is Representation,
            inherited=_RepresentationBaseType,
        )

        self.profiles = self.attr(
            "profiles",
            inherited=_RepresentationBaseType,
        )
        self.width = self.attr(
            "width",
            parser=int,
            inherited=_RepresentationBaseType,
        )
        self.height = self.attr(
            "height",
            parser=int,
            inherited=_RepresentationBaseType,
        )
        self.sar = self.attr(
            "sar",
            inherited=_RepresentationBaseType,
        )
        self.frameRate = self.attr(
            "frameRate",
            parser=MPDParsers.frame_rate,
            inherited=_RepresentationBaseType,
        )
        self.audioSamplingRate = self.attr(
            "audioSamplingRate",
            parser=int,
            inherited=_RepresentationBaseType,
        )
        self.codecs = self.attr(
            "codecs",
            inherited=_RepresentationBaseType,
        )
        self.scanType = self.attr(
            "scanType",
            inherited=_RepresentationBaseType,
        )

        self.contentProtections = self.children(ContentProtection)


class AdaptationSet(_RepresentationBaseType):
    __tag__ = "AdaptationSet"

    parent: Period

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.id = self.attr("id")
        self.group = self.attr("group")
        self.lang = self.attr("lang")
        self.contentType = self.attr("contentType")
        self.par = self.attr("par")
        self.minBandwidth = self.attr("minBandwidth", parser=int)
        self.maxBandwidth = self.attr("maxBandwidth", parser=int)
        self.minWidth = self.attr("minWidth", parser=int)
        self.maxWidth = self.attr("maxWidth", parser=int)
        self.minHeight = self.attr("minHeight", parser=int)
        self.maxHeight = self.attr("maxHeight", parser=int)
        self.minFrameRate = self.attr("minFrameRate", parser=MPDParsers.frame_rate)
        self.maxFrameRate = self.attr("maxFrameRate", parser=MPDParsers.frame_rate)
        self.segmentAlignment = self.attr(
            "segmentAlignment",
            parser=MPDParsers.bool_str,
            default=False,
        )
        self.subsegmentAlignment = self.attr(
            "subsegmentAlignment",
            parser=MPDParsers.bool_str,
            default=False,
        )
        self.subsegmentStartsWithSAP = self.attr(
            "subsegmentStartsWithSAP",
            parser=int,
            default=0,
        )
        self.bitstreamSwitching = self.attr(
            "bitstreamSwitching",
            parser=MPDParsers.bool_str,
        )

        self.baseURLs = self.children(BaseURL)
        self.segmentBase = self.only_child(SegmentBase, period=self.parent)
        self.segmentList = self.only_child(SegmentList, period=self.parent)
        self.segmentTemplate = self.only_child(SegmentTemplate, period=self.parent)
        self.representations = self.children(Representation, minimum=1, period=self.parent)


class Representation(_RepresentationBaseType):
    __tag__ = "Representation"

    parent: AdaptationSet
    mimeType: str

    def __init__(self, *args, period: Period, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.period = period

        self.id = self.attr(
            "id",
            required=True,
        )
        self.bandwidth = self.attr(
            "bandwidth",
            parser=MPDParsers.bandwidth,
            required=True,
        )

        self.ident = self.parent.parent.id, self.parent.id, self.id

        self.baseURLs = self.children(BaseURL)
        self.subRepresentations = self.children(SubRepresentation)
        self.segmentBase = self.only_child(SegmentBase, period=self.period)
        self.segmentList = self.only_child(SegmentList, period=self.period)
        self.segmentTemplate = self.only_child(SegmentTemplate, period=self.period)

    @property
    def lang(self):
        return self.parent.lang

    @property
    def bandwidth_rounded(self) -> float:
        return round(self.bandwidth, 1 - int(math.log10(self.bandwidth)))

    def segments(
        self,
        sequence: int = -1,
        init: bool = True,
        timestamp: datetime | None = None,
        **kwargs,
    ) -> Iterator[DASHSegment]:
        """
        Segments are yielded when they are available

        Segments appear on a timeline, for dynamic content they are only available at a certain time
        and sometimes for a limited time. For static content they are all available at the same time.

        :param sequence: Sequence number
        :param init: Yield the init segment and perform other initialization logic for dynamic manifests
        :param timestamp: Optional initial timestamp for syncing timelines of multiple substreams
        :param kwargs: extra args to pass to the segment template/list
        :return: yields Segments
        """

        # segmentBase = self.segmentBase or self.walk_back_get_attr("segmentBase")
        segmentList = self.segmentList or self.walk_back_get_attr("segmentList")
        segmentTemplate = self.segmentTemplate or self.walk_back_get_attr("segmentTemplate")

        if segmentTemplate:
            yield from segmentTemplate.segments(
                self.ident,
                self.base_url,
                init=init,
                timestamp=timestamp,
                RepresentationID=self.id,
                Bandwidth=int(self.bandwidth * 1000),
                **kwargs,
            )
        elif segmentList:
            yield from segmentList.segments(
                sequence=sequence,
                init=init,
            )
        else:
            yield DASHSegment(
                num=sequence,
                init=False,
                discontinuity=False,
                uri=self.base_url,
                duration=self.period.duration.total_seconds() or self.root.mediaPresentationDuration.total_seconds(),
                available_at=self.period.availabilityStartTime,
                byterange=None,
            )


class SubRepresentation(_RepresentationBaseType):
    __tag__ = "SubRepresentation"


class _SegmentBaseType(MPDNode):
    parent: Period | AdaptationSet | Representation

    _ancestors = (Period, AdaptationSet, Representation)

    def __init__(self, *args, period: Period, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.period = period

        self.timescale = self.attr(
            "timescale",
            parser=int,
            default=self._find_default("timescale", 1),
        )
        self.presentationTimeOffset = self.attr(
            "presentationTimeOffset",
            parser=MPDParsers.parse_timedelta(self.timescale),
            default=self._find_default("presentationTimeOffset", timedelta()),
        )
        self.availabilityTimeOffset = self.attr(
            "availabilityTimeOffset",
            parser=MPDParsers.parse_timedelta(self.timescale),
            default=self._find_default("availabilityTimeOffset", timedelta()),
        )

        self.initialization = self.only_child(Initialization) or self._find_default("initialization")

    @overload
    def _find_default(self, attr: str) -> Any | None: ...  # pragma: no cover

    @overload
    def _find_default(self, attr: str, default: TAttrDefault) -> TAttrDefault: ...  # pragma: no cover

    def _find_default(self, attr: str, default: TAttrDefault | None = None) -> Any | None:
        """Find default values from nodes of the same type on ancestor nodes"""
        # the node attribute on each ancestor is named after its node tag, with the first character being lowercase
        nodeattr = f"{self.__tag__[0].lower()}{self.__tag__[1:]}"
        # start with the parent node, to avoid an unnecessary failed lookup on the current node
        value = self.parent.walk_back_get_attr(
            attr,
            self._ancestors,
            lambda node: getattr(node, nodeattr, None),
        )
        return default if value is None else value


class _MultipleSegmentBaseType(_SegmentBaseType):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.duration = self.attr(
            "duration",
            parser=int,
            default=self._find_default("duration"),
        )
        self.startNumber = self.attr(
            "startNumber",
            parser=int,
            default=self._find_default("startNumber", 1),
        )

        self.duration_seconds = self.duration / self.timescale if self.duration else 0.0

        self.segmentTimeline = self.only_child(SegmentTimeline) or self._find_default("segmentTimeline")

    @staticmethod
    def make_url(base_url: str, url: str | None) -> str:
        if not url:  # pragma: no cover
            return base_url

        base_scheme = urlparse(base_url).scheme
        scheme = urlparse(url).scheme
        if is_insecure_scheme(base_scheme, scheme):
            raise MPDParsingError(f"Prevented access to insecure resource in manifest: {base_scheme=!r} {scheme=!r}")

        return urljoin(base_url, url)


class SegmentBase(_SegmentBaseType):
    __tag__ = "SegmentBase"


class SegmentList(_MultipleSegmentBaseType):
    __tag__ = "SegmentList"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.segmentURLs = self.children(SegmentURL)

    # noinspection PyUnusedLocal
    def segments(
        self,
        sequence: int = -1,
        init: bool = True,
    ) -> Iterator[DASHSegment]:
        if init and self.initialization:  # pragma: no branch
            yield DASHSegment(
                num=-1,
                init=True,
                discontinuity=False,
                uri=self.make_url(self.base_url, self.initialization.source_url),
                duration=0.0,
                available_at=self.period.availabilityStartTime,
                byterange=self.initialization.range,
            )
        for num, segment_url in self.segment_urls(sequence):
            yield DASHSegment(
                num=num,
                init=False,
                discontinuity=False,
                uri=self.make_url(self.base_url, segment_url.media),
                duration=self.duration_seconds,
                available_at=self.period.availabilityStartTime,
                byterange=segment_url.media_range,
            )

    def segment_urls(self, sequence: int) -> Iterator[tuple[int, SegmentURL]]:
        if sequence == -1:
            if self.root.type == "static":
                # yield all segments in a static manifest
                sequence = self.startNumber
                segment_urls = self.segmentURLs
            else:
                # yield a specific number of segments from the live-edge of dynamic manifests
                sequence = self.calculate_optimal_start()
                segment_urls = self.segmentURLs[sequence - self.startNumber :]

        else:
            # skip segments with a lower number than the remembered segment number
            # and check if we've skipped any segments after reloading the manifest
            offset = sequence - self.startNumber

            if offset >= 0:
                # no segments were skipped: yield a slice of the segments
                segment_urls = self.segmentURLs[offset:]
            else:
                # segments were skipped: yield all segments and set the correct segment number
                sequence = self.startNumber
                segment_urls = self.segmentURLs

        yield from enumerate(segment_urls, sequence)

    def calculate_optimal_start(self) -> int:
        """Calculate the optimal segment number to start based on the suggestedPresentationDelay"""
        suggested_delay = self.root.suggestedPresentationDelay

        if not self.duration:
            log.info(f"Unknown segment duration. Falling back to an offset of {MPD.DEFAULT_LIVE_EDGE_SEGMENTS} segments.")
            offset = MPD.DEFAULT_LIVE_EDGE_SEGMENTS
        else:
            offset = max(0, math.ceil(suggested_delay.total_seconds() / self.duration_seconds))

        start = self.startNumber + len(self.segmentURLs) - offset
        log.debug(f"Calculated optimal offset is {offset} segments. First segment is {start}.")

        return start


class SegmentTemplate(_MultipleSegmentBaseType):
    __tag__ = "SegmentTemplate"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.fmt_initialization = self.attr(
            "initialization",
            parser=MPDParsers.segment_template,
        )
        self.fmt_media = self.attr(
            "media",
            parser=MPDParsers.segment_template,
        )

    def segments(
        self,
        ident: TTimelineIdent,
        base_url: str,
        init: bool = True,
        timestamp: datetime | None = None,
        **kwargs,
    ) -> Iterator[DASHSegment]:
        if init:
            init_url = self.format_initialization(base_url, **kwargs)
            if init_url:  # pragma: no branch
                yield DASHSegment(
                    num=-1,
                    init=True,
                    discontinuity=False,
                    uri=init_url,
                    duration=0.0,
                    available_at=self.period.availabilityStartTime,
                    byterange=None,
                )
        for media_url, num, duration, available_at in self.format_media(ident, base_url, timestamp=timestamp, **kwargs):
            yield DASHSegment(
                num=num,
                init=False,
                discontinuity=False,
                uri=media_url,
                duration=duration,
                available_at=available_at,
                byterange=None,
            )

    def segment_numbers(self, timestamp: datetime | None = None) -> Iterator[tuple[int, datetime]]:
        """
        yield the segment number and when it will be available.

        There are two cases for segment number generation, "static" and "dynamic":

        In the case of static streams, the segment number starts at the startNumber and counts
        up to the number of segments that are represented by the periods-duration.

        In the case of dynamic streams, the segments should appear at the specified time.
        In the simplest case, the segment number is based on the time since the availabilityStartTime.
        """

        if not self.duration_seconds:  # pragma: no cover
            raise MPDParsingError("Unknown segment durations: missing duration/timescale attributes on SegmentTemplate")

        number_iter: Iterator[int] | Sequence[int]
        available_iter: Iterator[datetime]

        if self.root.type == "static":
            available_iter = repeat(self.period.availabilityStartTime)
            duration = self.period.duration.total_seconds() or self.root.mediaPresentationDuration.total_seconds()
            if duration:
                number_iter = range(self.startNumber, int(duration / self.duration_seconds) + 1)
            else:
                number_iter = count(self.startNumber)
        else:
            current_time = timestamp or now()
            since_start = current_time - self.period.availabilityStartTime - self.presentationTimeOffset

            suggested_delay = self.root.suggestedPresentationDelay
            buffer_time = self.root.minBufferTime

            # Segment number
            seconds_offset = (since_start - suggested_delay - buffer_time).total_seconds()
            number_offset = max(0, int(seconds_offset / self.duration_seconds))
            number_iter = count(self.startNumber + number_offset)

            # Segment availability time
            available_offset = timedelta(seconds=number_offset * self.duration_seconds)
            available_start = self.period.availabilityStartTime + available_offset
            available_iter = count_dt(
                available_start,
                timedelta(seconds=self.duration_seconds),
            )

            log.debug(f"Stream start: {self.period.availabilityStartTime}")
            log.debug(f"Current time: {current_time}")
            log.debug(f"Availability: {available_start}")
            log.debug(
                "; ".join([
                    f"presentationTimeOffset: {self.presentationTimeOffset}",
                    f"suggestedPresentationDelay: {self.root.suggestedPresentationDelay}",
                    f"minBufferTime: {self.root.minBufferTime}",
                ]),
            )
            log.debug(
                "; ".join([
                    f"segmentDuration: {self.duration_seconds}",
                    f"segmentStart: {self.startNumber}",
                    f"segmentOffset: {number_offset} ({seconds_offset}s)",
                ]),
            )

        yield from zip(number_iter, available_iter, strict=False)

    def segment_timeline(self, ident: TTimelineIdent) -> Iterator[tuple[int, TimelineSegment, datetime]]:
        if not self.segmentTimeline:  # pragma: no cover
            raise MPDParsingError("Missing SegmentTimeline in SegmentTemplate")

        if self.root.type == "static":
            yield from zip(count(self.startNumber), self.segmentTimeline.segments, repeat(self.period.availabilityStartTime))
        else:
            time = self.root.timelines[ident]
            is_initial = time == -1

            threshold = self.root.publishTime - self.root.suggestedPresentationDelay

            # transform the timeline into a segment list
            timeline = []
            available_at = self.root.publishTime

            # the last segment in the timeline is the most recent one
            # so, work backwards and calculate when each of the segments was
            # available, based on the durations relative to the publish-time
            for number, segment in reversed(list(zip(count(self.startNumber), self.segmentTimeline.segments))):
                # stop once the suggestedPresentationDelay is reached on the first manifest parsing
                # or when a segment with a lower or equal time value was already returned from an earlier manifest
                if is_initial and available_at <= threshold or segment.t <= time:
                    break

                timeline.append((number, segment, available_at))
                available_at -= timedelta(seconds=segment.d / self.timescale)

            # return the segments in chronological order
            for number, segment, available_at in reversed(timeline):
                self.root.timelines[ident] = segment.t
                yield number, segment, available_at

    def format_initialization(self, base_url: str, **kwargs) -> str | None:
        if self.fmt_initialization is not None:  # pragma: no branch
            return self.make_url(base_url, self.fmt_initialization(**kwargs))

    def format_media(
        self,
        ident: TTimelineIdent,
        base_url: str,
        timestamp: datetime | None = None,
        **kwargs,
    ) -> Iterator[tuple[str, int, float, datetime]]:
        if self.fmt_media is None:  # pragma: no cover
            return

        if not self.segmentTimeline:
            log.debug("Generating segment numbers for %s playlist: %r", self.root.type, ident)
            duration = self.duration_seconds
            for number, available_at in self.segment_numbers(timestamp=timestamp):
                url = self.make_url(base_url, self.fmt_media(Number=number, **kwargs))
                yield url, number, duration, available_at
        else:
            log.debug("Generating segment timeline for %s playlist: %r", self.root.type, ident)
            for number, segment, available_at in self.segment_timeline(ident):
                url = self.make_url(base_url, self.fmt_media(Time=segment.t, Number=number, **kwargs))
                duration = segment.d / self.timescale
                yield url, number, duration, available_at


class SegmentTimeline(MPDNode):
    __tag__ = "SegmentTimeline"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.timescale = self.walk_back_get_attr("timescale")

        self.timeline_segments = self.children(_TimelineSegment)

    @property
    def segments(self) -> Iterator[TimelineSegment]:
        t = 0
        for tsegment in self.timeline_segments:
            if t == 0 and tsegment.t is not None:
                t = tsegment.t
            # check the start time from MPD
            for _ in range(tsegment.r + 1):
                yield TimelineSegment(t, tsegment.d)
                t += tsegment.d


class _TimelineSegment(MPDNode):
    __tag__ = "S"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.t = self.attr("t", parser=int)
        self.d = self.attr("d", parser=int, required=True)
        self.r = self.attr("r", parser=int, default=0)


class Initialization(MPDNode):
    __tag__ = "Initialization"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.source_url = self.attr("sourceURL")
        self.range = self.attr(
            "range",
            parser=MPDParsers.range,
        )


class SegmentURL(MPDNode):
    __tag__ = "SegmentURL"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.media = self.attr("media")
        self.media_range = self.attr(
            "mediaRange",
            parser=MPDParsers.range,
        )


class ContentProtection(MPDNode):
    __tag__ = "ContentProtection"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.schemeIdUri = self.attr("schemeIdUri")
        self.value = self.attr("value")
        self.default_KID = self.attr("default_KID")



from __future__ import annotations

from collections import abc
from copy import copy, deepcopy
from functools import singledispatch
from re import Pattern
from typing import Any

from lxml.etree import Element, iselement

from streamlink.exceptions import PluginError
from streamlink.validate._exception import ValidationError
from streamlink.validate._schemas import (
    AllSchema,
    AnySchema,
    AttrSchema,
    GetItemSchema,
    ListSchema,
    NoneOrAllSchema,
    OptionalSchema,
    RegexSchema,
    TransformSchema,
    UnionGetSchema,
    UnionSchema,
    XmlElementSchema,
)


class Schema(AllSchema):
    """
    The base class for creating validation schemas.

    A wrapper for :class:`AllSchema <_schemas.AllSchema>` with a wrapper method for :func:`validate`
    which by default raises :class:`PluginError <streamlink.exceptions.PluginError>` on error.
    """

    # TODO: replace default PluginError exception
    def validate(self, value: Any, name: str = "result", exception: type[Exception] = PluginError) -> Any:
        try:
            return validate(self, value)
        except ValidationError as err:
            raise exception(f"Unable to validate {name}: {err}") from None


# ----


@singledispatch
def validate(schema, value):
    if schema != value:
        raise ValidationError(
            "{value} does not equal {expected}",
            value=repr(value),
            expected=repr(schema),
            schema="equality",
        )

    return value


@validate.register
def _validate_type(schema: type, value):
    if not isinstance(value, schema):
        raise ValidationError(
            "Type of {value} should be {expected}, but is {actual}",
            value=repr(value),
            expected=schema.__name__,
            actual=type(value).__name__,
            schema=type,
        )

    return value


# singledispatch doesn't support typing.Union/types.UnionType on py<311, so keep each register() call for now
@validate.register(list)
@validate.register(tuple)
@validate.register(set)
@validate.register(frozenset)
def _validate_sequence(schema: list | tuple | set | frozenset, value):
    cls = type(schema)
    validate(cls, value)
    any_schemas = AnySchema(*schema)

    return cls(validate(any_schemas, v) for v in value)


@validate.register
def _validate_dict(schema: dict, value):
    cls = type(schema)
    validate(cls, value)
    new = cls()

    for key, subschema in schema.items():
        if isinstance(key, OptionalSchema):
            if key.key not in value:
                continue
            key = key.key

        if type(key) in (type, AllSchema, AnySchema, TransformSchema, UnionSchema):
            for subkey, subvalue in value.items():
                try:
                    newkey = validate(key, subkey)
                except ValidationError as err:
                    raise ValidationError("Unable to validate key", schema=dict) from err
                try:
                    newvalue = validate(subschema, subvalue)
                except ValidationError as err:
                    raise ValidationError("Unable to validate value", schema=dict) from err
                new[newkey] = newvalue
            break

        if key not in value:
            raise ValidationError(
                "Key {key} not found in {value}",
                key=repr(key),
                value=repr(value),
                schema=dict,
            )

        try:
            new[key] = validate(subschema, value[key])
        except ValidationError as err:
            raise ValidationError("Unable to validate value of key {key}", key=repr(key), schema=dict) from err

    return new


@validate.register
def _validate_callable(schema: abc.Callable, value):
    if not schema(value):
        raise ValidationError(
            "{callable} is not true",
            callable=f"{getattr(schema, '__name__', schema.__class__.__name__)}({value!r})",
            schema=abc.Callable,
        )

    return value


@validate.register
def _validate_pattern(schema: Pattern, value):
    if not isinstance(value, (str, bytes)):
        raise ValidationError(
            "Type of {value} should be str or bytes, but is {actual}",
            value=repr(value),
            actual=type(value).__name__,
            schema=Pattern,
        )

    try:
        result = schema.search(value)
    except TypeError as err:
        raise ValidationError(err, schema=Pattern) from None

    return result


@validate.register
def _validate_allschema(schema: AllSchema, value):
    for subschema in schema.schema:
        value = validate(subschema, value)

    return value


@validate.register
def _validate_anyschema(schema: AnySchema, value):
    errors = []
    for subschema in schema.schema:
        try:
            return validate(subschema, value)
        except ValidationError as err:
            errors.append(err)

    raise ValidationError(*errors, schema=AnySchema)


@validate.register
def _validate_noneorallschema(schema: NoneOrAllSchema, value):
    if value is not None:
        try:
            for subschema in schema.schema:
                value = validate(subschema, value)
        except ValidationError as err:
            raise ValidationError(err, schema=NoneOrAllSchema) from None

    return value


@validate.register
def _validate_listschema(schema: ListSchema, value):
    if not isinstance(value, list):
        raise ValidationError(
            "Type of {value} should be list, but is {actual}",
            value=repr(value),
            actual=type(value).__name__,
            schema=ListSchema,
        )
    if len(value) != len(schema.schema):
        raise ValidationError(
            "Length of list ({length}) does not match expectation ({expected})",
            length=len(value),
            expected=len(schema.schema),
            schema=ListSchema,
        )

    new = []
    errors = []
    for k, v in enumerate(schema.schema):
        try:
            new.append(validate(v, value[k]))
        except ValidationError as err:
            errors.append(err)

    if errors:
        raise ValidationError(*errors, schema=ListSchema)

    return new


@validate.register
def _validate_regexschema(schema: RegexSchema, value):
    if not isinstance(value, (str, bytes)):
        raise ValidationError(
            "Type of {value} should be str or bytes, but is {actual}",
            value=repr(value),
            actual=type(value).__name__,
            schema=RegexSchema,
        )

    try:
        result = getattr(schema.pattern, schema.method)(value)
    except TypeError as err:
        raise ValidationError(err, schema=RegexSchema) from None

    if result is None:
        raise ValidationError(
            "Pattern {pattern} did not match {value}",
            pattern=repr(schema.pattern.pattern),
            value=repr(value),
            schema=RegexSchema,
        )

    return result


@validate.register
def _validate_transformschema(schema: TransformSchema, value):
    validate(abc.Callable, schema.func)
    return schema.func(value, *schema.args, **schema.kwargs)


@validate.register
def _validate_getitemschema(schema: GetItemSchema, value):
    item = schema.item if type(schema.item) is tuple and not schema.strict else (schema.item,)
    idx = 0
    key = None
    try:
        for key in item:
            if iselement(value):
                value = value.attrib[key]
            else:
                value = value[key]
            idx += 1
        return value
    except (KeyError, IndexError):
        # only return default value on last item in nested lookup
        if idx < len(item) - 1:
            raise ValidationError(
                "Item {key} was not found in object {value}",
                key=repr(key),
                value=repr(value),
                schema=GetItemSchema,
            ) from None
        return schema.default
    except (TypeError, AttributeError) as err:
        raise ValidationError(
            "Could not get key {key} from object {value}",
            key=repr(key),
            value=repr(value),
            schema=GetItemSchema,
        ) from err


@validate.register
def _validate_attrschema(schema: AttrSchema, value):
    new = copy(value)

    for key, subschema in schema.schema.items():
        if not hasattr(value, key):
            raise ValidationError(
                "Attribute {key} not found on object {value}",
                key=repr(key),
                value=repr(value),
                schema=AttrSchema,
            )

        try:
            value = validate(subschema, getattr(value, key))
        except ValidationError as err:
            raise ValidationError(
                "Could not validate attribute {key}",
                key=repr(key),
                schema=AttrSchema,
            ) from err

        setattr(new, key, value)

    return new


@validate.register
def _validate_xmlelementschema(schema: XmlElementSchema, value):
    validate(iselement, value)
    tag = value.tag
    attrib = value.attrib
    text = value.text
    tail = value.tail

    if schema.tag is not None:
        try:
            tag = validate(schema.tag, value.tag)
        except ValidationError as err:
            raise ValidationError("Unable to validate XML tag", schema=XmlElementSchema) from err

    if schema.attrib is not None:
        try:
            attrib = validate(schema.attrib, dict(value.attrib))
        except ValidationError as err:
            raise ValidationError("Unable to validate XML attributes", schema=XmlElementSchema) from err

    if schema.text is not None:
        try:
            text = validate(schema.text, value.text)
        except ValidationError as err:
            raise ValidationError("Unable to validate XML text", schema=XmlElementSchema) from err

    if schema.tail is not None:
        try:
            tail = validate(schema.tail, value.tail)
        except ValidationError as err:
            raise ValidationError("Unable to validate XML tail", schema=XmlElementSchema) from err

    new = Element(tag, attrib)
    new.text = text
    new.tail = tail
    for child in value:
        new.append(deepcopy(child))

    return new


@validate.register
def _validate_uniongetschema(schema: UnionGetSchema, value):
    return schema.seq(validate(getter, value) for getter in schema.getters)


@validate.register
def _validate_unionschema(schema: UnionSchema, value):
    try:
        return validate_union(schema.schema, value)
    except ValidationError as err:
        raise ValidationError("Could not validate union", schema=UnionSchema) from err


# ----


# noinspection PyUnusedLocal
@singledispatch
def validate_union(schema, value):
    raise ValidationError(
        "Invalid union type: {type}",
        type=type(schema).__name__,
    )


@validate_union.register
def _validate_union_dict(schema: dict, value):
    new = type(schema)()
    for key, subschema in schema.items():
        is_optional = isinstance(key, OptionalSchema)
        if is_optional:
            key = key.key

        try:
            new[key] = validate(subschema, value)
        except ValidationError as err:
            if is_optional:
                continue

            raise ValidationError(
                "Unable to validate union {key}",
                key=repr(key),
                schema=dict,
            ) from err

    return new


# singledispatch doesn't support typing.Union/types.UnionType on py<311, so keep each register() call for now
@validate_union.register(list)
@validate_union.register(tuple)
@validate_union.register(set)
@validate_union.register(frozenset)
def _validate_union_sequence(schemas: list | tuple | set | frozenset, value):
    return type(schemas)(validate(schema, value) for schema in schemas)



"""
$description European public service channel promoting culture, including magazine shows, concerts and documentaries.
$url arte.tv
$type live, vod
$metadata id
$metadata title
"""

import re
from operator import itemgetter

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    name="live",
    pattern=re.compile(
        r"https?://(?:\w+\.)?arte\.tv/(?P<language>[a-z]{2})/(?:direct|live)/?",
    ),
)
@pluginmatcher(
    name="vod",
    pattern=re.compile(
        r"https?://(?:\w+\.)?arte\.tv/(?:guide/)?(?P<language>[a-z]{2})/(?:videos/)?(?P<video_id>(?!RC-|videos)[^/]+?)/.+",
    ),
)
class ArteTV(Plugin):
    API_URL = "https://api.arte.tv/api/player/v2/config/{language}/{id}"

    def _get_streams(self):
        self.id = self.match["video_id"] if self.matches["vod"] else "LIVE"

        json_url = self.API_URL.format(
            language=self.match["language"],
            id=self.id,
        )
        streams, metadata = self.session.http.get(
            json_url,
            schema=validate.Schema(
                validate.parse_json(),
                {"data": {"attributes": dict}},
                validate.get(("data", "attributes")),
                {
                    "streams": validate.any(
                        [],
                        [
                            validate.all(
                                {
                                    "slot": int,
                                    "protocol": str,
                                    "url": validate.url(),
                                },
                                validate.union_get("slot", "protocol", "url"),
                            ),
                        ],
                    ),
                    "metadata": {
                        "title": str,
                        "subtitle": validate.any(None, str),
                    },
                },
                validate.union_get("streams", "metadata"),
            ),
        )

        self.title = f"{metadata['title']} - {metadata['subtitle']}" if metadata["subtitle"] else metadata["title"]

        for _slot, protocol, url in sorted(streams, key=itemgetter(0)):
            if "HLS" not in protocol:
                continue
            return HLSStream.parse_variant_playlist(self.session, url)


__plugin__ = ArteTV



"""
$description Vietnamese live TV channels owned by the People's Committee of Ho Chi Minh City.
$url htv.com.vn
$type live
$region Vietnam
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.times import localnow


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?htv\.com\.vn/truc-tuyen(?:\?channel=(?P<channel>\w+)&?|$)"),
)
class HTV(Plugin):
    def get_channels(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath(".//*[contains(@class,'channel-list')]//a[@data-id][@data-code]"),
                [
                    validate.union_get("data-id", "data-code"),
                ],
            ),
        )

        return dict(data)

    def _get_streams(self):
        channels = self.get_channels()

        if not channels:
            log.error("No channels found")
            return

        log.debug(f"channels={channels}")

        channel_id = self.match.group("channel")
        if channel_id is None:
            channel_id, channel_code = next(iter(channels.items()))
        elif channel_id in channels:
            channel_code = channels[channel_id]
        else:
            log.error(f"Unknown channel ID: {channel_id}")
            return

        log.info(f"Channel: {channel_code}")

        json = self.session.http.post(
            "https://www.htv.com.vn/HTVModule/Services/htvService.aspx",
            data={
                "method": "GetScheduleList",
                "channelid": channel_id,
                "template": "AjaxSchedules.xslt",
                "channelcode": channel_code,
                "date": localnow().strftime("%d-%m-%Y"),
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "success": bool,
                    "chanelUrl": validate.url(),
                },
            ),
        )

        if not json["success"]:
            log.error("API error: success not true")
            return

        hls_url = self.session.http.get(
            json["chanelUrl"],
            headers={"Referer": self.url},
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[contains(text(), 'playlist.m3u8')]/text()"),
                validate.none_or_all(
                    re.compile(r"""var\s+iosUrl\s*=\s*(?P<q>")(?P<url>.+?)(?P=q)"""),
                    validate.none_or_all(
                        validate.get("url"),
                        validate.url(),
                    ),
                ),
            ),
        )

        if hls_url:
            return HLSStream.parse_variant_playlist(
                self.session,
                hls_url,
                headers={"Referer": "https://hplus.com.vn/"},
            )


__plugin__ = HTV
